"""Cell geometry model classes and routines."""
from math import cos, sin
from typing import Iterator, List, Tuple

import numpy as np

from ..geometry import (
    add_empty_third_dimension,
    circle_segment,
    line,
    parabolic_deformation,
    rotate,
    rotate3d,
    rotate_and_mesh,
    shift,
)
from ..typing import DefaultsType

CircleType = Tuple[float, Tuple[float, float]]


class Shape:
    """Base class for implementing cell shapes."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict()

    def raw_points(self, simplify: bool = False) -> np.ndarray:
        pass

    def get_approximation_circles(self) -> Iterator[CircleType]:
        pass


class Shape3D(Shape):
    """Base class for implementing 3D cell shapes."""

    def raw_points3d(self, steps: int = 32, simplify: bool = False) -> np.ndarray:
        pass


class RodShaped(Shape):
    """Rod shaped cell geometry."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(length=2.0, width=1.0)

    def rod_raw_points(
        self, simplify: bool = False
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        diameter = self.width
        radius = diameter / 2.0
        length = self.length - diameter
        half_length = length / 2.0

        upper = line(
            [half_length, radius],
            [-half_length, radius],
            times=None if not simplify else 3,
        )
        lower = line(
            [-half_length, -radius],
            [half_length, -radius],
            times=None if not simplify else 3,
        )

        circle_left = circle_segment(radius, 90, 270, times=None if not simplify else 5)
        circle_left[:, 0] -= half_length

        circle_right = circle_segment(
            radius, -90, 90, times=None if not simplify else 5
        )
        circle_right[:, 0] += half_length

        return lower, circle_right, upper, circle_left

    def raw_points(self, simplify: bool = False) -> np.ndarray:
        lower, circle_right, upper, circle_left = self.rod_raw_points(simplify=simplify)
        return np.r_[lower, circle_right, upper, circle_left]

    def get_approximation_circles(self) -> Iterator[Tuple[float, Tuple[float, float]]]:
        diameter = self.width
        radius = diameter / 2.0
        length = self.length - diameter
        half_length = length / 2.0

        times = max(1, 2 * int(length / radius))

        for x in np.linspace(-half_length, half_length, times):
            yield radius, (x, 0)


class Rectangle(Shape):
    """Rectangular cell geometry."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(length=2.0, width=1.0)

    def raw_points(self, simplify: bool = False) -> DefaultsType:
        half_width, half_length = self.width / 2.0, self.length / 2.0

        return np.r_[
            line(
                [+half_length, +half_width],
                [-half_length, +half_width],
                times=None if not simplify else 3,
            ),
            line(
                [-half_length, +half_width],
                [-half_length, -half_width],
                times=None if not simplify else 3,
            ),
            line(
                [-half_length, -half_width],
                [+half_length, -half_width],
                times=None if not simplify else 3,
            ),
            line(
                [+half_length, -half_width],
                [+half_length, +half_width],
                times=None if not simplify else 3,
            ),
        ]

    def get_approximation_circles(self) -> Iterator[CircleType]:
        # not properly implemented
        diameter = self.width
        radius = diameter / 2.0
        length = self.length - diameter
        half_length = length / 2.0

        times = 2 * int(length / radius)

        for x in np.linspace(-half_length, half_length, times):
            yield radius, (x, 0)


class Square(Rectangle):
    """Square cell geometry."""

    def raw_points(self, simplify: bool = False) -> np.ndarray:
        # noinspection PyAttributeOutsideInit
        self.width = self.length
        return super().raw_points(simplify=simplify)


class BentRod(RodShaped):
    """Bent rod shaped cell geometry."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(bend_overall=0.0, bend_upper=0.0, bend_lower=0.0)

    def bend(self, points: np.ndarray) -> np.ndarray:
        u_idx, l_idx = points[:, 1] > 0, points[:, 1] < 0

        points[u_idx, :] = parabolic_deformation(points[u_idx, :], self.bend_upper)
        points[l_idx, :] = parabolic_deformation(points[l_idx, :], self.bend_lower)

        points = parabolic_deformation(points, self.bend_overall)

        return points

    def raw_points(self, simplify: bool = False) -> np.ndarray:
        lower, circle_right, upper, circle_left = self.rod_raw_points(simplify=simplify)

        points = np.r_[lower, circle_right, upper, circle_left]

        points = self.bend(points)

        return points

    def raw_points3d(
        self, steps: int = 16, simplify: bool = False
    ) -> Tuple[np.ndarray, np.ndarray]:
        lower, circle_right, upper, circle_left = self.rod_raw_points(simplify=simplify)
        points = np.r_[lower, circle_right, upper, circle_left]

        points, tris = rotate_and_mesh(add_empty_third_dimension(points), steps=steps)

        points = self.bend(points)

        return points, tris

    def get_approximation_circles(self) -> Iterator[CircleType]:
        radii = []
        offsets = []
        for radius, offset in super().get_approximation_circles():
            radii.append(radius)
            offsets.append(offset)
        offsets = np.array(offsets)
        offsets = parabolic_deformation(offsets, self.bend_overall)

        for radius, offset in zip(radii, offsets):
            yield radius, offset


class Coccoid(Shape):
    """Coccoid (spherical) cell geometry."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(length=1.0)

    def raw_points(self, simplify: bool = False) -> np.ndarray:
        radius = self.length / 2

        circle_left = circle_segment(radius, 90, 270, times=None if not simplify else 5)
        circle_right = circle_segment(
            radius, -90, 90, times=None if not simplify else 5
        )

        return np.r_[circle_right, circle_left]

    def get_approximation_circles(self) -> Iterator[CircleType]:
        yield self.length / 2, (0.0, 0.0)


class Ellipsoid(Coccoid):
    """Ellipsoid cell geometry."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(length=2.0, width=1.0)

    def raw_points(self, simplify: bool = False) -> np.ndarray:
        points = super().raw_points()

        a = self.length / 2
        b = self.width / 2

        points[:, 1] *= b / a

        return points


class WithPosition:
    """Mixin adding a cell position."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(position=lambda: [0.0, 0.0])


class WithAngle:
    """Mixin adding a cell angle."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(angle=0.0)


class WithFluorescence:
    """Mixin adding a fluorescence value."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(fluorescences=lambda: [0.0])


class WithProperDivisionBehavior:
    """Mixin adding division angle calculation."""

    def get_division_positions(self, count: int = 2) -> List[List[float]]:
        # must have a length, a position and an angle

        sin_a, cos_a = sin(self.angle), cos(self.angle)

        x, y = self.position

        return [
            [float(x + factor * cos_a), float(y + factor * sin_a)]
            for factor in np.linspace(
                -self.length / 2 / 2, self.length / 2 / 2, num=count
            )
        ]


class AutoMesh3D(Shape3D):
    """Mixin adding automatic solid-of-revolution generation."""

    def points3d_on_canvas(
        self, steps: int = 16, simplify: bool = False
    ) -> Tuple[np.ndarray, np.ndarray]:
        points, triangles = self.raw_points3d(steps=steps, simplify=simplify)

        axis_vector = ((0,), (0,), (1,))

        position3d = self.position

        return shift(rotate3d(points, self.angle, axis_vector), position3d), triangles

    def raw_points3d(
        self, steps: int = 16, simplify: bool = False
    ) -> Tuple[np.ndarray, np.ndarray]:
        return rotate_and_mesh(
            add_empty_third_dimension(self.raw_points(simplify=simplify)), steps=steps
        )


class CellGeometry(WithAngle, WithPosition, AutoMesh3D):
    """Cell geometry base by combining multiple mixins."""

    def points_on_canvas(self) -> np.ndarray:
        return shift(rotate(self.raw_points(), self.angle), self.position)


__all__ = [
    'Shape',
    'Shape3D',
    'RodShaped',
    'Rectangle',
    'Square',
    'BentRod',
    'Coccoid',
    'Ellipsoid',
    'WithPosition',
    'WithAngle',
    'WithFluorescence',
    'WithProperDivisionBehavior',
    'AutoMesh3D',
    'CellGeometry',
]



"""Output using matplotlib."""
import warnings
from typing import Tuple

from matplotlib import pyplot
from tunable import Tunable

from ..parameters import Height, Width
from ..simulation.simulator import World
from . import Output, check_overwrite, ensure_path_and_extension_and_number


class MicrometerPerCm(Tunable):
    default: float = 2.5


class PlotRenderer(Output, Output.Default):
    """Output using matplotlib."""

    def __init__(self):
        super().__init__()

        self.fig = self.ax = None

    def output(self, world: World, **kwargs) -> Tuple[pyplot.Figure, pyplot.Axes]:
        if self.fig is None:
            self.fig = pyplot.figure(
                figsize=(
                    Width.value / MicrometerPerCm.value / 2.51,
                    Height.value / MicrometerPerCm.value / 2.51,
                )
            )

        fig = self.fig

        if self.ax is None:
            self.ax = fig.add_subplot(111)

        ax = self.ax

        ax.clear()

        for cell in world.cells:
            points = cell.points_on_canvas()
            ax.plot(points[:, 0], points[:, 1])

        for boundary in world.boundaries:
            for start, stop in zip(boundary, boundary[1:]):
                ax.plot([start[0], stop[0]], [start[1], stop[1]])

        ax.set_xlim(0, Width.value)
        ax.set_ylim(0, Height.value)
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            pyplot.tight_layout()

        return fig, ax

    def write(
        self,
        world: World,
        file_name: str,
        output_count: int = 0,
        overwrite: bool = False,
        **kwargs
    ) -> None:
        fig, ax = self.output(world)

        extensions = ['.png'] + [
            '.' + extension
            for extension in fig.canvas.get_supported_filetypes().keys()
            if extension != 'png'
        ]

        fig.savefig(
            check_overwrite(
                ensure_path_and_extension_and_number(
                    file_name, extensions, output_count
                ),
                overwrite=overwrite,
            )
        )

    def display(self, world: World, **kwargs) -> None:
        if self.fig:
            if not pyplot.fignum_exists(self.fig.number):
                raise KeyboardInterrupt

        pyplot.ion()
        self.output(world)
        pyplot.show()
        self.fig.canvas.draw()
        self.fig.canvas.flush_events()


__all__ = ['PlotRenderer', 'MicrometerPerCm']



"""CLI entrypoint."""
import importlib
import logging
import sys
from argparse import ArgumentParser, Namespace
from pathlib import Path
from types import ModuleType
from typing import Callable, Iterable, Optional

from tunable import TunableSelectable

from ..output import all as output_all
from ..random import RRF
from . import render, simulate, training

_ = output_all

log = logging.getLogger(__name__)


def load_class_from_module(module_class: str, default_class_name: str) -> type:
    if ':' in module_class:
        splits = module_class.split(':')
        module_name, class_name = ':'.join(splits[:-1]), splits[-1]
    else:
        module_name, class_name = module_class, default_class_name

    if Path(module_name).exists():
        module = ModuleType(default_class_name)
        exec(Path(module_name).read_text(), module.__dict__)
    else:
        module = importlib.import_module(module_name)

    class_type = getattr(module, class_name)
    return class_type


def parse_arguments_and_init(
    args: Iterable[str],
    parser_callback: Optional[Callable[[ArgumentParser], None]] = None,
) -> Namespace:
    """
    Basic setup (i.e. logging) and argument parsing.

    :param args: Arguments
    :param parser_callback: Additional callback to configure the argument parser
    :return: Parsed arguments
    """
    logging.basicConfig(
        level=logging.INFO,
        format="%(asctime)-15s.%(msecs)03d %(name)s %(levelname)s %(message)s",
        datefmt='%Y-%m-%d %H:%M:%S',
    )

    parser = ArgumentParser()

    parser.add_argument('-o', '--output-file', dest='output', default=None)
    parser.add_argument(
        '-w', '--overwrite', dest='overwrite', default=False, action='store_true'
    )
    parser.add_argument(
        '-p', '--prefix', dest='prefix', default=False, action='store_true'
    )
    parser.add_argument('-c', '--cell', dest='cell', default='cellsium.cli:SizerCell')
    verbose_group = parser.add_mutually_exclusive_group()
    verbose_group.add_argument(
        '-q', '--quiet', dest='quiet', default=False, action='store_true'
    )
    verbose_group.add_argument(
        '-v', '--verbose', dest='verbose', default=1, action='count'
    )

    TunableSelectable.setup_and_parse(parser, args=args)

    if parser_callback:
        parser_callback(parser)

    parsed_args = parser.parse_args(args=args)

    if parsed_args.quiet:
        log.setLevel(logging.WARNING)
    elif parsed_args.verbose == 1:
        log.setLevel(logging.INFO)
    elif parsed_args.verbose > 1:
        log.setLevel(logging.DEBUG)

    if parsed_args.cell:
        parsed_args.cell = load_class_from_module(
            parsed_args.cell, default_class_name='Cell'
        )

    return parsed_args


subcommands = {simulate, render, training}
subcommand_default = simulate


def main(args: Optional[Iterable[str]] = None) -> Optional[int]:
    """
    Main entrypoint of the script, will redirect to various sub-scripts.

    :param args: arguments, if not specified will be taken from sys.argv
    :return: The return code of the individual subcommand
    """
    if args is None:
        args = sys.argv[1:]

    if not args:
        subcommand = subcommand_default
    else:
        subcommands_dict = {
            subcommand.__name__.split('.')[-1]: subcommand for subcommand in subcommands
        }
        first_arg = args[0]
        if first_arg in subcommands_dict:
            subcommand = subcommands_dict[first_arg]
            del args[0]
        else:
            subcommand = subcommand_default

    subcommand_argparser = getattr(subcommand, 'subcommand_argparser', None)

    parsed_args = parse_arguments_and_init(
        args=args, parser_callback=subcommand_argparser
    )

    seed = RRF.seed()
    log.info("Seeding with %s" % (seed,))

    return subcommand.subcommand_main(args=parsed_args)



from typing import Any

import numpy as np
from tunable import Selectable, Tunable

from .. import BaseSimulator


class PlacementSimulationSimplification(Tunable):
    """How much the placement should be simplified,
    0: use the normal shapes,
    1: use simplified shapes,
    2: use many-circle approximations"""

    default: int = 0


class PlacementSimulation(BaseSimulator, Selectable):
    pass


class PhysicalPlacement(PlacementSimulation, PlacementSimulation.Virtual):
    def __init__(self):

        self.cell_bodies = {}
        self.cell_shapes = {}

        self.boundaries = []

    def clear(self) -> None:
        for cell in list(self.cell_bodies.keys()):
            self.remove(cell)

    def _get_positions(self) -> np.ndarray:
        array = np.zeros((len(self.cell_bodies), 3))
        for n, body in enumerate(
            sorted(self.cell_bodies.values(), key=lambda body_: id(body_))
        ):
            array[n, :] = body.position[0], body.position[1], body.angle
        return array

    @staticmethod
    def _all_distances(before: np.ndarray, after: np.ndarray) -> np.ndarray:
        return np.sqrt(((after - before) ** 2).sum(axis=1))

    @classmethod
    def _total_distance(cls, before: np.ndarray, after: np.ndarray) -> float:
        return cls._all_distances(before, after).sum()

    @classmethod
    def _mean_distance(cls, before: np.ndarray, after: np.ndarray) -> float:
        return cls._all_distances(before, after).mean()


class NoPlacement(PlacementSimulation):
    pass


def ensure_python(value: Any) -> Any:
    if isinstance(value, np.ndarray):
        return value.tolist()
    else:
        return value



"""CLI package, home to the individual entry points"""
from pathlib import Path
from typing import Any, Optional

from ..model import PlacedCell, SizerCell, TimerCell
from ..output import Output
from ..simulation.placement import PlacementSimulation
from ..simulation.simulator import Simulator


class TimerCell(PlacedCell, TimerCell):
    """Cell."""

    pass


class SizerCell(PlacedCell, SizerCell):
    """Cell."""


Cell = SizerCell


def initialize_simulator() -> Simulator:
    """
    Constructor helper for a simulator.

    :return: Simulator instance
    """
    simulator = Simulator()
    ps = PlacementSimulation()

    simulator.sub_simulators += [ps]

    return simulator


def initialize_cells(
    simulator: Simulator,
    count: int = 1,
    cell_type: Optional[PlacedCell] = None,
    sequence: Any = None,
) -> Simulator:
    """
    Initialize cells and add them to a simulator.

    :param simulator: Simulator to add cells to.
    :param count: Count of cells to generate
    :param cell_type: cell type to use
    :param sequence: Random number sequence to use
    :return: Simulator
    """
    if cell_type is None:
        cell_type = Cell

    random_sequences = cell_type.get_random_sequences(sequence=sequence)

    for _ in range(count):

        init_kwargs = {k: next(v) for k, v in random_sequences.items()}

        # handling for items which are generated jointly,
        # but need to go to separate kwargs
        for k, v in list(init_kwargs.items()):
            if "__" in k:
                del init_kwargs[k]
                k_fragments = k.split("__")
                for inner_k, inner_v in zip(k_fragments, v):
                    init_kwargs[inner_k] = inner_v

        cell = cell_type(**init_kwargs)
        cell.birth()
        simulator.add(cell)

    return simulator


def add_output_prefix(output_name: str, output: Output) -> str:
    """
    Adds an prefix to an output filename.

    :param output_name: Output name
    :param output: Output object
    :return: Name
    """
    output_name = Path(output_name)

    output_name = output_name.parent / (
        output.__class__.__name__ + "-" + output_name.name
    )

    return str(output_name)



"""Placement simulation using Pymunk physics engine."""
import numpy as np
from tunable import Tunable

from ...model import PlacedCell
from .base import (
    PhysicalPlacement,
    PlacementSimulation,
    PlacementSimulationSimplification,
    ensure_python,
)

try:
    import pymunkoptions

    pymunkoptions.options['debug'] = False
except ImportError:
    pymunkoptions = None  # PyMunk 6.0.0 has removed pymunkoptions
# noinspection PyPep8
import pymunk


class ChipmunkPlacementRadius(Tunable):
    """Chipmunk placement radius, additional radius objects will have around them"""

    default: float = 0.05


class Chipmunk(PhysicalPlacement, PlacementSimulation, PlacementSimulation.Default):

    verbose: bool = False
    look_back_threshold: int = 5
    convergence_check_interval: int = 15

    def __init__(self):
        self.space = pymunk.Space(threaded=False)

        # self.space.threads = 2
        self.space.iterations = 100

        self.space.gravity = 0, 0

        self.boundary_bodies = []
        self.boundary_segments = []

        super().__init__()

    def add_boundary(self, coordinates: np.ndarray) -> None:
        coordinates = np.array(coordinates)

        boundary_body = pymunk.Body(body_type=pymunk.Body.STATIC)

        boundary_segments = []

        for start, stop in zip(coordinates, coordinates[1:]):
            self.boundaries.append([start, stop])
            segment = pymunk.Segment(
                boundary_body, ensure_python(start), ensure_python(stop), 0.0
            )
            boundary_segments.append(segment)

        self.space.add(boundary_body, *boundary_segments)

        self.boundary_bodies.append(boundary_body)
        self.boundary_segments.append(boundary_segments)

    def add(self, cell: PlacedCell) -> None:
        body = pymunk.Body(1.0, 1.0)
        body.position = pymunk.Vec2d(cell.position[0], cell.position[1])
        body.angle = cell.angle

        if PlacementSimulationSimplification.value == 2:
            shapes = tuple(
                pymunk.Circle(body, float(radius), offset=ensure_python(offset))
                for radius, offset in cell.get_approximation_circles()
            )
        else:
            points = cell.raw_points(
                simplify=PlacementSimulationSimplification.value == 1
            )

            poly = pymunk.Poly(body, ensure_python(points))
            poly.unsafe_set_radius(ChipmunkPlacementRadius.value)

            shapes = (poly,)

        self.cell_bodies[cell] = body
        self.cell_shapes[cell] = shapes

        self.space.add(body, *shapes)

    def remove(self, cell: PlacedCell) -> None:
        self.space.remove(self.cell_bodies[cell], *self.cell_shapes[cell])

        del self.cell_bodies[cell]
        del self.cell_shapes[cell]

    def clear(self) -> None:
        super().clear()

        for boundary_body, boundary_segments in zip(
            self.boundary_bodies, self.boundary_segments
        ):
            self.space.remove(boundary_body)
            self.space.remove(*boundary_segments)

        self.boundary_bodies.clear()
        self.boundary_segments.clear()

    def step(self, timestep: float) -> None:
        if len(self.cell_bodies) == 0:
            return

        resolution = 0.1 * 10
        times = timestep / resolution
        last = self.inner_step(
            time_step=resolution, iterations=int(times), epsilon=1e-12
        )
        _ = last

    def inner_step(
        self,
        time_step: float = 0.1,
        iterations: int = 9999,
        converge: bool = True,
        epsilon: float = 0.1,
    ) -> float:

        first_positions = self._get_positions()[:, :2]

        if converge:
            self._inner_step_step_attempt_converge(
                epsilon, first_positions, iterations, time_step
            )
        else:
            self._inner_step_step_ignore_converge(iterations, time_step)

        after_positions = self._get_positions()[:, :2]

        self._inner_step_update_positions()

        return self._total_distance(first_positions, after_positions)

    def _inner_step_update_positions(self) -> None:
        for cell, body in self.cell_bodies.items():
            cell.position = [body.position[0], body.position[1]]
            cell.angle = body.angle

    def _inner_step_step_ignore_converge(
        self, iterations: int, time_step: float
    ) -> None:
        for _ in range(iterations):
            self.space.step(time_step)

    def _inner_step_step_attempt_converge(
        self,
        epsilon: float,
        first_positions: np.ndarray,
        iterations: int,
        time_step: float,
    ) -> None:
        converging = False
        look_back = 0
        convergence_check = self.convergence_check_interval
        before_positions = first_positions.copy()
        for _ in range(iterations):
            self.space.step(time_step)

            convergence_check -= 1

            if convergence_check > 0:
                continue

            convergence_check = self.convergence_check_interval

            after_positions = self._get_positions()[:, :2]

            dist = (
                self._mean_distance(before_positions, after_positions)
                * time_step
                * self.convergence_check_interval
            )

            before_positions[:] = after_positions

            if dist < epsilon:
                look_back += 1
            else:
                look_back = 0

            if look_back > self.look_back_threshold:
                if self.verbose:
                    print("Stopping due to look back threshold.")
                break

            if self.verbose:
                print(_, dist, look_back)

            if not converging:
                if dist > 0:
                    converging = True
            else:
                if dist < epsilon:
                    break

            before_positions[:] = after_positions



"""Output as Trackmate XML lineage files, compatible with the JuNGLE extensions."""
# noinspection PyPep8Naming
import xml.etree.ElementTree as ET
from pathlib import Path

from tunable import Tunable

from ..parameters import Calibration, Height, Width, um_to_pixel
from ..simulation.simulator import World
from . import Output, check_overwrite, ensure_path_and_extension


class TrackMateXMLExportFluorescences(Tunable):
    """Names for the fluorescences for the JuNGLE TrackMate format"""

    # default = ''
    default: str = 'Crimson,YFP'


class TrackMateXMLExportLengthTypo(Tunable):
    """Whether to output in the classic JuNGLE TrackMate format"""

    default: bool = True


EMPTY_TRACKMATE_XML = """<?xml version="1.0" encoding="UTF-8"?>
<TrackMate version="2.7.3">
  <Settings>
    <ImageData filename="FILENAME" folder="" width="1" height="1" nslices="1" nframes="2" pixelwidth="1.0" 
    pixelheight="1.0" voxeldepth="1.0" timeinterval="1.0" />
    <BasicSettings xstart="0" xend="1" ystart="0" yend="1" zstart="0" zend="0" tstart="0" tend="2" />
<InitialSpotFilter feature="QUALITY" value="0.0" isabove="true" />
    <DetectorSettings DETECTOR_NAME="OVERLAY_DETECTOR" TARGET_CHANNEL="1" />
    <SpotFilterCollection />
    <TrackerSettings TRACKER_NAME="SPARSE_LAP_TRACKER" CUTOFF_PERCENTILE="0.9"
    ALTERNATIVE_LINKING_COST_FACTOR="1.05" BLOCKING_VALUE="Infinity">
      <Linking LINKING_MAX_DISTANCE="15.0">
        <FeaturePenalties />
      </Linking>
      <GapClosing ALLOW_GAP_CLOSING="true" GAP_CLOSING_MAX_DISTANCE="6.0" MAX_FRAME_GAP="2">
        <FeaturePenalties />
      </GapClosing>
      <TrackSplitting ALLOW_TRACK_SPLITTING="true" SPLITTING_MAX_DISTANCE="10.0">
        <FeaturePenalties />
      </TrackSplitting>
      <TrackMerging ALLOW_TRACK_MERGING="false" MERGING_MAX_DISTANCE="15.0">
        <FeaturePenalties />
      </TrackMerging>
    </TrackerSettings>
    <TrackFilterCollection />
    <AnalyzerCollection>
      <SpotAnalyzers>
        <Analyzer key="SPOT_FLUORESCENCE_ANALYZER" />
        <Analyzer key="SPOT_MEASUREMENT_ANALYZER" />
        <Analyzer key="MANUAL_SPOT_COLOR_ANALYZER" />
        <Analyzer key="Spot descriptive statistics" />
        <Analyzer key="Spot radius estimator" />
        <Analyzer key="Spot contrast and SNR" />
      </SpotAnalyzers>
      <EdgeAnalyzers>
        <Analyzer key="Edge target" />
        <Analyzer key="Edge mean location" />
        <Analyzer key="Edge velocity" />
        <Analyzer key="MANUAL_EDGE_COLOR_ANALYZER" />
      </EdgeAnalyzers>
      <TrackAnalyzers>
        <Analyzer key="Branching analyzer" />
        <Analyzer key="Track duration" />
        <Analyzer key="Track index" />
        <Analyzer key="Track location" />
        <Analyzer key="Velocity" />
        <Analyzer key="TRACK_SPOT_QUALITY" />
      </TrackAnalyzers>
    </AnalyzerCollection>
  </Settings>
  <Model spatialunits="pixels" timeunits="frames">
    <FeatureDeclarations>
      <SpotFeatures>
        <Feature feature="QUALITY" 
        name="Quality" shortname="Quality" dimension="QUALITY" isint="false" />
        <Feature feature="POSITION_X" 
        name="X" shortname="X" dimension="POSITION" isint="false" />
        <Feature feature="POSITION_Y" 
        name="Y" shortname="Y" dimension="POSITION" isint="false" />
        <Feature feature="POSITION_Z" 
        name="Z" shortname="Z" dimension="POSITION" isint="false" />
        <Feature feature="POSITION_T" 
        name="T" shortname="T" dimension="TIME" isint="false" />
        <Feature feature="FRAME" 
        name="Frame" shortname="Frame" dimension="NONE" isint="true" />
        <Feature feature="RADIUS" 
        name="Radius" shortname="R" dimension="LENGTH" isint="false" />
        <Feature feature="VISIBILITY" 
        name="Visibility" shortname="Visibility" dimension="NONE" isint="true" />
        <Feature feature="PIXELS" 
        name="Pixels" shortname="Pixels" dimension="INTENSITY" isint="false" />
        <Feature feature="AREA" 
        name="Cell area" shortname="Area" dimension="INTENSITY_SQUARED" isint="false" />
        <Feature feature="LENGHT" 
        name="Cell length" shortname="Length" dimension="LENGTH" isint="false" />
      </SpotFeatures>
      <EdgeFeatures>
        <Feature feature="SPOT_SOURCE_ID" name="Source spot ID" shortname="Source ID" dimension="NONE" isint="true" />
        <Feature feature="SPOT_TARGET_ID" name="Target spot ID" shortname="Target ID" dimension="NONE" isint="true" />
        <Feature feature="LINK_COST" name="Link cost" shortname="Cost" dimension="NONE" isint="false" />
      </EdgeFeatures>
      <TrackFeatures>
        <Feature feature="TRACK_INDEX" name="Track index" shortname="Index" dimension="NONE" isint="true" />
        <Feature feature="TRACK_ID" name="Track ID" shortname="ID" dimension="NONE" isint="true" />
      </TrackFeatures>
    </FeatureDeclarations>
    <AllSpots nspots="0">
    </AllSpots>
    <AllTracks>
    </AllTracks>
    <FilteredTracks>
    </FilteredTracks>
  </Model>
  <GUIState state="ConfigureViews">
    <View key="HYPERSTACKDISPLAYER" />
  </GUIState>
</TrackMate>
"""  # noqa


class TrackMateXML(Output):
    def __init__(self):
        super().__init__()

        self.root = ET.fromstring(EMPTY_TRACKMATE_XML)

        self.xml = ET.ElementTree(self.root)

        data = self.root.find('Settings/ImageData').attrib
        data['width'] = str(int(um_to_pixel(Width.value)))
        data['height'] = str(int(um_to_pixel(Height.value)))
        data['pixelwidth'] = str(Calibration.value)
        data['pixelheight'] = str(Calibration.value)

        self.image_data = data

        settings = self.root.find('Settings/BasicSettings').attrib

        settings['xend'] = str(int(um_to_pixel(Width.value)))
        settings['yend'] = str(int(um_to_pixel(Height.value)))

        spot_features = self.root.find('Model/FeatureDeclarations/SpotFeatures')

        for f in TrackMateXMLExportFluorescences.value.split(','):
            if not f:
                continue
            ET.SubElement(spot_features, 'Feature').attrib.update(
                dict(
                    feature='%s_FLUORESCENCE_MEAN' % f.upper(),
                    name='%s Mean' % f,
                    shortname='%s' % f,
                    dimension='INTENSITY',
                    isint='false',
                )
            )

            ET.SubElement(spot_features, 'Feature').attrib.update(
                dict(
                    feature='%s_FLUORESCENCE_STDDEV' % f.upper(),
                    name='%s StdDev' % f,
                    shortname='%s SD' % f,
                    dimension='INTENSITY',
                    isint='false',
                )
            )

            ET.SubElement(spot_features, 'Feature').attrib.update(
                dict(
                    feature='%s_FLUORESCENCE_TOTAL' % f.upper(),
                    name='%s Total' % f,
                    shortname='%sTot' % f,
                    dimension='INTENSITY',
                    isint='false',
                )
            )

        if not TrackMateXMLExportLengthTypo.value:
            for feature in spot_features:
                if feature.attrib['feature'] == 'LENGHT':
                    feature.attrib['feature'] = 'LENGTH'

        self.basic_settings = settings

        self.frame_counter = 0

        self.all_spots = self.root.find('Model/AllSpots')
        self.spot_counter = 0

        self.cell_to_spot = {}
        self.spot_to_cell = {}
        self.id_to_cell = {}

        self.tracks = {}
        self.track_counter = 0

        self.all_tracks = self.root.find('Model/AllTracks')
        self.filtered_tracks = self.root.find('Model/FilteredTracks')

    def output(self, world: World, time: float = 0.0, **kwargs) -> None:
        frame_counter = self.frame_counter

        self.frame_counter += 1

        self.image_data['nframes'] = str(frame_counter)
        self.basic_settings['tend'] = str(frame_counter)

        # frame
        group = ET.SubElement(self.all_spots, 'SpotsInFrame')
        group.attrib['frame'] = str(frame_counter)

        def add_track():
            elem = ET.SubElement(self.all_tracks, 'Track')
            self.track_counter += 1
            track_count = self.track_counter
            elem.attrib['name'] = 'Track_' + str(track_count)
            elem.attrib['TRACK_INDEX'] = str(track_count)
            elem.attrib['TRACK_ID'] = str(track_count)

            the_elem = elem

            elem = ET.SubElement(self.filtered_tracks, 'TrackID')
            elem.attrib['TRACK_ID'] = str(track_count)

            return the_elem

        def connect(track, from_, to_):
            elem = ET.SubElement(track, 'Edge')

            elem.attrib['SPOT_SOURCE_ID'] = str(from_)
            elem.attrib['SPOT_TARGET_ID'] = str(to_)

            elem.attrib['LINK_COST'] = str(1.0)

        old_cell_to_spot = self.cell_to_spot.copy()
        old_id_to_cell = self.id_to_cell.copy()

        for cell in world.cells:
            self.spot_counter += 1
            self.cell_to_spot[cell] = self.spot_counter
            self.spot_to_cell[self.spot_counter] = cell
            self.id_to_cell[cell.id_] = cell

        # # when dealing with local history sequences
        # def sub_sequences(seq):
        #     for n in range(len(seq)):
        #         yield seq[n:]

        for cell in world.cells:
            spot = ET.SubElement(group, 'Spot')

            # history = tuple(cell.lineage_history[::-1])
            # this will probably not work if more than one division
            # is between xml-write intervals

            if cell in self.tracks:
                connect(
                    self.tracks[cell], old_cell_to_spot[cell], self.cell_to_spot[cell]
                )
            else:
                if (
                    cell.parent_id in old_id_to_cell
                    and old_id_to_cell[cell.parent_id] in self.tracks
                ):
                    connect(
                        self.tracks[old_id_to_cell[cell.parent_id]],
                        old_cell_to_spot[old_id_to_cell[cell.parent_id]],
                        self.cell_to_spot[cell],
                    )
                    self.tracks[cell] = self.tracks[old_id_to_cell[cell.parent_id]]
                else:
                    self.tracks[cell] = add_track()

            spot.attrib['ID'] = str(self.cell_to_spot[cell])
            spot.attrib['name'] = str(self.cell_to_spot[cell])

            for f in TrackMateXMLExportFluorescences.value.split(','):
                if not f:
                    continue

                spot.attrib['%s_FLUORESCENCE_TOTAL' % f.upper()] = '0.0'
                spot.attrib['%s_FLUORESCENCE_MEAN' % f.upper()] = '0.0'
                spot.attrib['%s_FLUORESCENCE_STDDEV' % f.upper()] = '0.0'

            spot.attrib['AREA'] = str(0.0)
            spot.attrib['VISIBILITY'] = str(1)

            spot.attrib['POSITION_T'] = str(time)  # minutes or seconds?

            if not TrackMateXMLExportLengthTypo.value:
                spot.attrib['LENGTH'] = str(cell.length)
            else:
                spot.attrib['LENGHT'] = str(cell.length)  # sic!

            spot.attrib['PIXELS'] = str(0.0)

            spot.attrib['POSITION_X'] = str(um_to_pixel(cell.position[0]))
            spot.attrib['POSITION_Y'] = str(
                um_to_pixel(Height.value - cell.position[1])
            )
            spot.attrib['POSITION_Z'] = str(0.0)

            spot.attrib['FRAME'] = str(frame_counter)

            spot.attrib['RADIUS'] = str(5.0)
            spot.attrib['QUALITY'] = str(1.0)

        self.all_spots.attrib['nspots'] = str(self.spot_counter)

    # noinspection PyMethodOverriding
    def write(
        self,
        world: World,
        file_name: str,
        time: float = 0.0,
        overwrite: bool = False,
        **kwargs
    ) -> None:
        self.output(world, time=time)

        self.image_data['filename'] = Path(
            ensure_path_and_extension(file_name, '.tif')
        ).name

        file_name = check_overwrite(
            ensure_path_and_extension(file_name, '.xml'), overwrite=overwrite
        )

        self.xml.write(file_name)

    def display(self, world: World, **kwargs) -> None:
        raise RuntimeError('Unsupported')


__all__ = ['TrackMateXML']



from . import Output
from .gt import COCOOutput, GenericMaskOutput, YOLOOutput
from .mesh import MeshOutput
from .plot import PlotRenderer
from .render import (
    FluorescenceRenderer,
    NoisyUnevenIlluminationPhaseContrast,
    PhaseContrastRenderer,
    PlainRenderer,
    TiffOutput,
    UnevenIlluminationPhaseContrast,
)
from .serialization import CsvOutput, JsonPickleSerializer, QuickAndDirtyTableDumper
from .svg import SvgRenderer
from .xml import TrackMateXML

__all__ = [
    'Output',
    'YOLOOutput',
    'COCOOutput',
    'GenericMaskOutput',
    'MeshOutput',
    'PlotRenderer',
    'PlainRenderer',
    'FluorescenceRenderer',
    'PhaseContrastRenderer',
    'UnevenIlluminationPhaseContrast',
    'NoisyUnevenIlluminationPhaseContrast',
    'TiffOutput',
    'JsonPickleSerializer',
    'QuickAndDirtyTableDumper',
    'CsvOutput',
    'SvgRenderer',
    'TrackMateXML',
]



import fnmatch
import gc
import os
import sys
from contextlib import contextmanager

import pytest
from tunable import Selectable, TunableManager

from ..cli import initialize_cells, initialize_simulator
from ..model import (
    BentRod,
    Coccoid,
    Ellipsoid,
    IdCounter,
    Rectangle,
    RodShaped,
    SizerCell,
    Square,
    WithFluorescence,
    generate_cell,
)
from ..model.initialization import RandomFluorescence
from ..random import RRF
from .generate_dxf_file import generate_dxf_file


@contextmanager
def _patch_sys_modules(remove=None):
    gc.collect()

    if remove is None:
        remove = []

    sys_modules_keys = set()
    sys_modules_extract = {}

    for key, value in sys.modules.items():
        for to_remove in remove:
            if fnmatch.fnmatch(key, to_remove):
                sys_modules_extract[key] = value

        sys_modules_keys.add(key)

    try:
        for key in sys_modules_extract.keys():
            del sys.modules[key]

        yield

    finally:
        new_sys_modules_keys = set(sys.modules.keys())
        difference = new_sys_modules_keys - sys_modules_keys

        for key in difference:
            del sys.modules[key]

        # important for proper cleanup
        for key in sys_modules_extract.keys():
            if key in sys.modules:
                del sys.modules[key]

        gc.collect()

        for key, value in sys_modules_extract.items():
            sys.modules[key] = value


@pytest.fixture
def patch_sys_modules():
    return _patch_sys_modules


def get_recursive_subclasses(class_):
    def collect(inner_class, collector=set(), skip=None):
        if inner_class != skip:
            collector.add(inner_class)
        for iter_class in inner_class.__subclasses__():
            collect(iter_class)
        return collector

    return collect(class_, skip=class_)


def fix_bases(base_class, pattern='*', keep=None):
    for item in get_recursive_subclasses(base_class):
        if keep and item in keep:
            continue
        if fnmatch.fnmatch(item.__name__, pattern):
            new_bases = tuple([base for base in item.__bases__ if base != base_class])
            # if not new_bases:
            #     new_bases = (object,)

            item.__bases__ = new_bases


@contextmanager
def _fix_dynamic_module_reloading_subclass_problems():
    keep_selectable = get_recursive_subclasses(Selectable)
    # keep_tunable = get_recursive_subclasses(Tunable)

    try:
        yield
    finally:
        fix_bases(Selectable, '*', keep=keep_selectable)
        # fix_bases(Tunable, '*', keep=keep_tunable)
        gc.collect()


@pytest.fixture
def fix_dynamic_module_reloading_subclass_problems():
    return _fix_dynamic_module_reloading_subclass_problems


@contextmanager
def _mock_sys_argv(new_value):
    argv = sys.argv
    try:
        sys.argv = new_value
        yield
    finally:
        sys.argv = argv


@pytest.fixture
def mock_sys_argv():
    return _mock_sys_argv


def perform_reset_state():
    # we have implicit state in CellSium in three places ...

    # ... the tunables / selectables

    TunableManager.init()

    Selectable.SelectableChoice.overrides.clear()
    Selectable.SelectableChoice.parameters.clear()

    # ... the random number generator

    RRF.seed(1)

    # ... the id counter to assign globally unique ids to cells

    IdCounter.reset()


@pytest.fixture
def reset_state():
    perform_reset_state()

    return perform_reset_state


@pytest.fixture
def simulator(reset_state):
    simulator = initialize_simulator()

    initialize_cells(simulator, count=1, sequence=RRF.sequence)

    simulator.simulation.world.commit()

    return simulator


@pytest.fixture
def add_cell_zoo():
    def _inner(simulator):

        cell_types = [
            BentRod,
            RodShaped,
            Coccoid,
            Ellipsoid,
            Square,
            Rectangle,
            (BentRod, WithFluorescence, RandomFluorescence),
        ]

        for cell_type in cell_types:
            if not isinstance(cell_type, tuple):
                cell_type = (cell_type,)

            cell_type += (SizerCell,)

            generated_cell_type = generate_cell(
                *cell_type, name='Placed' + cell_type[0].__name__
            )

            initialize_cells(simulator, count=1, cell_type=generated_cell_type)

        simulator.simulation.world.commit()

    return _inner


@pytest.fixture
def dxf_file(tmpdir):
    location = tmpdir.join('test.dxf')

    def _callable(line_type):
        generate_dxf_file(str(location), line_type=line_type)
        return location

    return _callable


@contextmanager
def _tunables(*tunables_list):
    old_values = {}

    for t, v in tunables_list:
        old_values[t] = t.value

    try:
        for t, v in tunables_list:
            assert t.test(v)
            t.value = v

        yield

    finally:
        for t, v in tunables_list:
            t.value = old_values[t]
            t.reset()


@pytest.fixture
def tunables():
    return _tunables


@contextmanager
def _chdir(new_dir):
    cwd = os.getcwd()
    try:
        os.chdir(new_dir)
        yield
    finally:
        os.chdir(cwd)


@pytest.fixture
def chdir():
    return _chdir


@pytest.fixture
def nop(*args, **kwargs):
    def _nop(*args, **kwargs):
        pass

    return _nop



"""Placement simulation using Box2D physics engine."""
# noinspection PyPep8Naming
import Box2D as B2D
import numpy as np

from ...model import PlacedCell
from .base import (
    PhysicalPlacement,
    PlacementSimulation,
    PlacementSimulationSimplification,
    ensure_python,
)


class Box2D(PhysicalPlacement, PlacementSimulation):

    verbose: bool = False

    def __init__(self):
        self.world = B2D.b2World()
        self.world.gravity = 0, 0

        super().__init__()

    def add_boundary(self, coordinates: np.ndarray) -> None:
        coordinates = np.array(coordinates)

        for start, stop in zip(coordinates, coordinates[1:]):
            self.boundaries.append([start, stop])
            self.world.CreateStaticBody(
                shapes=B2D.b2EdgeShape(
                    vertices=[ensure_python(start), ensure_python(stop)]
                )
            )

    def add(self, cell: PlacedCell) -> None:
        # Box2D only allows for 16 vertices per body,
        # and the current implementation would vastly exceed that
        # hence only simplifications
        assert PlacementSimulationSimplification.value != 0

        if PlacementSimulationSimplification.value == 2:
            shapes = [
                B2D.b2CircleShape(pos=offset, radius=radius)
                for radius, offset in cell.get_approximation_circles()
            ]
        else:
            shapes = [
                B2D.b2PolygonShape(
                    vertices=cell.raw_points(
                        simplify=PlacementSimulationSimplification.value == 1
                    ).tolist()
                )
            ]

        body = self.world.CreateDynamicBody(
            position=cell.position, angle=cell.angle, shapes=shapes
        )

        self.cell_bodies[cell] = body

    def remove(self, cell: PlacedCell) -> None:
        self.world.DestroyBody(self.cell_bodies[cell])

        del self.cell_bodies[cell]

    def step(self, timestep: float) -> None:
        velocity_iter, position_iter = 6, 30
        self.world.Step(timestep, velocity_iter, position_iter)

        for cell, body in self.cell_bodies.items():
            cell.position = body.position[0], body.position[1]
            cell.angle = float(body.angle)



"""The output package contains the various output modules."""
from pathlib import Path
from typing import Any, Optional, Tuple

from tunable import Selectable, Tunable

from ..simulation.simulator import World

ShapeType = Tuple[int, int]


def ensure_path(path: str) -> str:
    """
    Ensures that the parent directory to the to path exists.

    :param path: Path
    :return: the path
    """

    path = Path(path)
    if not path.parent.is_dir():
        path.parent.mkdir(parents=True, exist_ok=True)

    return str(path)


def ensure_extension(path: str, extension: str) -> str:
    """
    Ensures that the path ends with extension, possibly adding it.

    :param path: Path
    :param extension: Extension
    :return: Final path
    """
    path = Path(path)

    if not isinstance(extension, list):
        extension = [extension]

    if not path.suffix or path.suffix and path.suffix not in extension:
        path = path.parent / (path.name + extension[0])

    path = str(path)

    if OutputIndividualFilesWildcard.value in path:
        path = path.replace(OutputIndividualFilesWildcard.value, "")

    return path


def ensure_path_and_extension(path: str, extension: str) -> str:
    """
    Ensures that the parent directory to path exists,
    and it has extension, possibly by adding it.

    :param path: Path
    :param extension: Extension
    :return: Final path
    """
    ensure_path(path)
    return ensure_extension(path, extension)


def ensure_number(path: str, number: int, disable_individual: bool = False) -> str:
    """
    Depending on configuration, add a number to the path for consecutive output files.

    :param path: Path
    :param number: Number
    :param disable_individual: Possibility to disable adding of a number
    :return: Path with number
    """
    if OutputIndividualFiles.value and not disable_individual and number != -1:
        path = Path(path)

        stem = path.stem
        if OutputIndividualFilesWildcard.value not in stem:
            stem += OutputIndividualFilesWildcard.value

        digits = OutputIndividualFilesZeros.value

        stem = stem.replace(OutputIndividualFilesWildcard.value, f"{number:0>{digits}}")

        path = path.parent / (stem + path.suffix)

        path = str(path)

    return path


def ensure_path_and_extension_and_number(
    path: str, extension: str, number: int, disable_individual: bool = False
) -> str:
    """
    Ensures that a path exists, has an extension and a number.

    :param path: Path
    :param extension: Extension
    :param number: Number
    :param disable_individual: Whether to disable adding of number
    :return: Final path
    """
    path = ensure_number(path, number, disable_individual=disable_individual)
    return ensure_path_and_extension(path, extension)


def check_overwrite(path: str, overwrite: bool = False) -> str:
    """
    Check if a path exists, if so raising a RuntimeError if overwriting is disabled.

    :param path: Path
    :param overwrite: Whether to overwrite
    :return: Path
    """
    if Path(path).is_file() and not overwrite:
        raise RuntimeError(
            f"Requested existing {path!r} as output, but overwriting is disabled."
        )

    return path


class OutputIndividualFiles(Tunable):
    """Output individual files"""

    default: bool = True


class OutputIndividualFilesZeros(Tunable):
    """Amount of digits used for outputting the frame number of individual file names"""

    default: int = 3


class OutputIndividualFilesWildcard(Tunable):
    """Pattern for individual file names"""

    default: str = '{}'


class OutputReproducibleFiles(Tunable):
    """Output files in a reproducible manner"""

    default: bool = True


class Output(Selectable, Selectable.Multiple):
    """
    Base class of the Output classes.
    """

    def output(self, world: World, **kwargs) -> Optional[Any]:
        """
        Outputs the World, this function is usually called by either write or display.

        :param world: World
        :param kwargs: Additional arguments
        :return:
        """
        pass

    def write(self, world: World, file_name: str, **kwargs) -> None:
        """
        Output and write the World to file_name.

        :param world: World
        :param file_name: Filename to write output to
        :param kwargs: Additional arguments
        :return:
        """
        pass

    def display(self, world: World, **kwargs) -> None:
        """
        Output and display the World, e.g. via a GUI window.

        :param world: World
        :param kwargs: Additional arguments
        :return:
        """
        raise RuntimeError("Not implemented")



"""Cell model classes and routines, general."""
from copy import deepcopy
from typing import Any, List

from ..random import RRF
from ..typing import DefaultsType


def iter_through_class_hierarchy(cls: type) -> List[type]:
    """
    Iterate thru a class hierarchy and return all bases.

    :param cls: Type
    :return: List of bases
    """
    collector = []

    def _inner(cls_):
        if cls_ == object:
            return
        for base in cls_.__bases__:
            collector.append(base)
            _inner(base)

    _inner(cls)
    return collector


class InitializeWithParameters:
    """Mixin for objects with defaults."""

    def __init__(self, **kwargs):
        for cls_ in iter_through_class_hierarchy(self.__class__):
            if hasattr(cls_, 'defaults'):
                for k, v in cls_.defaults().items():
                    if hasattr(v, '__call__'):
                        v = v()
                    setattr(self, k, v)

        for k, v in kwargs.items():
            setattr(self, k, v)


class WithRandomSequences:
    """Mixin for objects with random sequences."""

    all_random_sequences_generated_for = {}

    @classmethod
    def get_random_sequences(cls, sequence: Any = None) -> Any:
        if sequence is not None and cls in cls.all_random_sequences_generated_for:
            del cls.all_random_sequences_generated_for[cls]

        if cls not in cls.all_random_sequences_generated_for:
            all_random_sequences = {}
            if sequence is None:
                sequence = RRF.sequence
            for cls_ in iter_through_class_hierarchy(cls):
                if hasattr(cls_, 'random_sequences'):
                    for key, value in cls_.random_sequences(sequence).items():
                        all_random_sequences[key] = value

            cls.all_random_sequences_generated_for[cls] = all_random_sequences

        return cls.all_random_sequences_generated_for[cls]

    @property
    def random(self) -> Any:
        class _Proxy:
            def __init__(self, backing):
                self.backing = backing

            def __getattr__(self, item):
                return self.backing[item]

        return _Proxy(self.get_random_sequences())


class Copyable:
    """Mixin for copyable objects."""

    def copy(self) -> "Copyable":
        return deepcopy(self)


class Representable:
    """Mixins for adding a repr implementation."""

    def __repr__(self):
        return (
            self.__class__.__name__
            + '('
            + ', '.join(['%s=%r' % (k, v) for k, v in sorted(self.__dict__.items())])
            + ')'
        )


class IdCounter:
    """Id provider singleton class."""

    id_counter: int = 0

    @classmethod
    def next_cell_id(cls) -> int:
        cls.id_counter += 1
        return cls.id_counter

    @classmethod
    def reset(cls) -> None:
        cls.id_counter = 0


class WithLineage:
    """Mixin providing lineage tracking."""

    def copy(self) -> "WithLineage":
        copy = super().copy()
        copy.next_cell_id()
        return copy

    # noinspection PyAttributeOutsideInit
    def next_cell_id(self) -> None:
        self.id_ = IdCounter.next_cell_id()

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(id_=lambda: IdCounter.next_cell_id(), parent_id=0)


class WithLineageHistory:
    """Mixin providing lineage history."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(lineage_history=lambda: [0])


class WithTemporalLineage:
    """Mixing providing temporal lineage history."""

    @staticmethod
    def defaults() -> DefaultsType:
        return dict(birth_time=0.0)


__all__ = [
    'InitializeWithParameters',
    'WithRandomSequences',
    'Copyable',
    'Representable',
    'IdCounter',
    'WithLineage',
    'WithLineageHistory',
    'WithTemporalLineage',
]



"""
The tests modules contains a doctests launcher.
"""

import doctest
from types import ModuleType


def collect_modules_recursive(start, blacklist=None):
    """
    Collects all modules and submodules in a recursive manner.

    :param start: the top module to start from
    :param blacklist: a string or list of strings of module
        (sub)names which should be ignored.
    :return:
    """

    if blacklist is None:
        blacklist = []
    elif not isinstance(blacklist, list):
        blacklist = [blacklist]

    collector = set()

    def _inner(current):
        if current in collector or current.__name__ in blacklist:
            return

        collector.add(current)

        for another in dir(current):
            another = getattr(current, another)
            if isinstance(another, ModuleType):
                if another.__name__.startswith(start.__name__):
                    _inner(another)

    _inner(start)

    return list(sorted(collector, key=lambda module: module.__name__))


def run_tests_recursively(start_module, blacklist=None, quiet=False):
    """
    Runs doctests recursively.

    :param start_module: the top module to start from
    :param blacklist: a string or list of strings of module
        (sub)names which should be ignored.
    :param quiet: whether to print infos about tests
    :return:
    """
    total_failures, total_tests = 0, 0
    for a_module in collect_modules_recursive(start_module, blacklist):
        failures, tests = doctest.testmod(a_module)
        total_failures += failures
        total_tests += tests

    if not quiet:
        print("Run %d tests in total." % (total_tests,))

    if total_failures > 0:
        if not quiet:
            print("Test failures occurred, exiting with non-zero status.")

    return total_failures


# noinspection PyUnresolvedReferences
def run_doctests(quiet=False):
    """
    Runs all the doctests.

    """

    import cellsium.cli.cli
    import cellsium.cli.render
    import cellsium.cli.training
    import cellsium.geometry
    import cellsium.model
    import cellsium.model.agent
    import cellsium.model.geometry
    import cellsium.output
    import cellsium.output.all
    import cellsium.parameters
    import cellsium.random
    import cellsium.simulation
    import cellsium.simulation.placement.base
    import cellsium.simulation.placement.pybox2d
    import cellsium.simulation.placement.pymunk
    import cellsium.simulation.simulator

    return run_tests_recursively(cellsium, quiet=quiet)


def main():
    return run_doctests()



"""Placement simulation package, contains the placement simulators."""
from .pymunk import Chipmunk

try:
    from .pybox2d import Box2D
except ImportError:
    Box2D = None
from .base import PlacementSimulation

__all__ = ['Chipmunk', 'Box2D', 'PlacementSimulation']



"""Simulation CLI entrypoint."""
import logging
import sys
from argparse import Namespace
from functools import partial, reduce
from time import time
from typing import Callable, Iterable, Iterator, Optional, Tuple, TypeVar

import numpy as np
from tunable import Tunable

from ...output import Output
from ...parameters import NewCellCount, h_to_s, s_to_h
from ...simulation.simulator import Simulator, Timestep, World
from .. import add_output_prefix, initialize_cells, initialize_simulator


class SimulationDuration(Tunable):
    """Time (simulated) the simulation should run"""

    default: float = 12.0


class SimulationOutputInterval(Tunable):
    """Time intervals (simulated) at which an output should be written"""

    default: float = 0.25


class SimulationTimestep(Tunable):
    """Time step at which the simulation state should be calculated"""

    default: float = 1.0 / 60.0


class SimulationOutputFirstState(Tunable):
    """Whether to output the first state"""

    default: bool = False


class BoundariesFile(Tunable):
    """Boundaries file (in DXF format) to add boundaries/geometrical constraints"""

    default: str = ""


class BoundariesScaleFactor(Tunable):
    """Scale factor for the boundaries"""

    default: float = 1.0


log = logging.getLogger(__name__)


def compose(*args):
    """
    Compose callable args.

    :param args: Args
    :return: Callable
    """
    return partial(reduce, lambda arg, fun: fun(arg), args)


def add_boundaries_from_dxf(
    file_name: str, simulator: Simulator, scale_factor: float = 1.0
) -> None:
    """
    Add boundaries from a DXF file.
    Supported are LWPolyline and Polyline objects.

    :param file_name: dxf file name
    :param simulator: Simulator instance to add the boundaries to
    :param scale_factor: Scale factor for the geometry
    :return: None
    """
    import ezdxf
    from ezdxf.entities import LWPolyline, Polyline

    dxf = ezdxf.readfile(file_name)

    for item in dxf.modelspace():
        points = None
        if isinstance(item, LWPolyline):
            points = np.array(list(item.get_points()))[:, :2]
        elif isinstance(item, Polyline):
            points = np.array(list(item.points()))[:, :2]
        else:
            log.warning("Warning, unknown type: %r", item.dxftype())

        if points is not None:
            points *= scale_factor

            simulator.add_boundary(points)


def add_boundaries_from_tunables(simulator: Simulator) -> Simulator:
    """
    Add boundaries to a simulator if the appropriate tunables are set.

    :param simulator: Simulator
    :return: Simulator
    """
    if BoundariesFile.value != '':
        add_boundaries_from_dxf(
            BoundariesFile.value, simulator, scale_factor=BoundariesScaleFactor.value
        )
    return simulator


def prepare_output_name(output_name: str, output: Output, prefix: str) -> str:
    """
    Prepare an output name.

    :param output_name: Output name
    :param output: Output object
    :param prefix: Prefix
    :return: Output name
    """
    if prefix:
        output_name = add_output_prefix(output_name, output=output)
    return output_name


def perform_outputs(
    world: World,
    simulation_time: float,
    outputs: Iterable[Output],
    output_name: Optional[str] = None,
    overwrite: bool = False,
    prefix: bool = False,
    output_count: int = 0,
) -> None:
    """
    Performs the output operations configured.

    :param world: World to output
    :param simulation_time: Simulation timepoint
    :param outputs: Outputs
    :param output_name: Name to output to
    :param overwrite: Whether to overwrite
    :param prefix: Whether to prefix the outputs with the name of the Output type
    :param output_count: The count of already outputted timesteps
    :return: None
    """
    log.debug("Outputting simulation state at %.2f h" % (s_to_h(simulation_time),))
    for output in outputs:
        output_before = time()

        if output_name:
            output.write(
                world,
                prepare_output_name(output_name, output, prefix),
                time=simulation_time,
                output_count=output_count,
                overwrite=overwrite,
            )
        else:
            output.display(world)

        output_after = time()

        log.debug(
            "Output %s took %.2fs"
            % (output.__class__.__name__, output_after - output_before)
        )


def initialize_output_times_from_tunables():
    """
    Initialize the duration, output_interval and last_output variables,
    from tunables.

    :return: duration, output_interval, last_output
    """
    duration = (
        h_to_s(SimulationDuration.value)
        if SimulationDuration.value > 0
        else float('inf')
    )

    output_interval = h_to_s(SimulationOutputInterval.value)

    if not SimulationOutputFirstState.value:
        last_output = 0.0
    else:
        last_output = -(h_to_s(SimulationOutputInterval.value) + sys.float_info.epsilon)

    return duration, output_interval, last_output


T = TypeVar('T')


def measure_duration(iterator: Iterator[T]) -> Iterator[Tuple[float, T]]:
    """
    Measure the (wall clock) time it takes to step forward the iterator,
    yielding the duration and iterator value.

    :param iterator: Iterator
    :return: Iterator of duration and iterator value
    """
    before = time()
    for value in iterator:
        after = time()
        yield (after - before), value
        before = time()


def perform_simulation(
    simulator: Optional[Simulator] = None,
    setup: Optional[Callable[[Simulator], Simulator]] = None,
    time_step: float = 1.0,
) -> Iterator[Timestep]:
    """
    Simple entrypoint to perform simulations.
    Will either accept a simulator instance or create one,
    call the setup callback on it if set,
    and then yield timesteps time_step apart indefinitely.

    :param simulator: Simulator
    :param setup: Setup callable
    :param time_step: Time step
    :return: Iterator of Timestep instances
    """
    if simulator is None:
        simulator = initialize_simulator()

    if setup:
        setup(simulator)

    while True:
        yield simulator.step(time_step)


def subcommand_main(args: Namespace) -> None:
    """
    Entry point for the 'simulate' subcommand.

    :param args: pre-parsed arguments
    :return: None
    """

    outputs = Output.SelectableGetMultiple()

    total_before = time()

    duration, output_interval, last_output = initialize_output_times_from_tunables()

    output_count = 0

    if duration == float('inf'):
        log.info("Simulation running in infinite mode ... press Ctrl-C to abort.")

    simulation_iterator = perform_simulation(
        setup=compose(
            add_boundaries_from_tunables,
            partial(initialize_cells, count=NewCellCount.value, cell_type=args.cell),
        ),
        time_step=h_to_s(SimulationTimestep.value),
    )

    interrupted = False
    try:
        for step_duration, ts in measure_duration(simulation_iterator):

            if ts.simulation.time > duration:
                break

            log.info(
                "Timestep took %.2fs, simulated time: %.2f h"
                % (step_duration, s_to_h(ts.simulation.time))
            )

            if (ts.simulation.time - last_output) >= output_interval > 0:
                last_output = ts.simulation.time

                perform_outputs(
                    ts.world,
                    ts.simulation.time,
                    outputs,
                    args.output,
                    overwrite=args.overwrite,
                    prefix=args.prefix,
                    output_count=output_count,
                )

                output_count += 1

    except KeyboardInterrupt:
        log.info("Ctrl-C pressed, stopping simulation.")
        interrupted = True

    total_after = time()
    log.info(
        "%s simulation took %.2fs"
        % (("Whole" if not interrupted else "Interrupted"), total_after - total_before)
    )



"""Type hints definitions."""

from typing import Any, Callable, Dict, Protocol, Union

DefaultsType = Dict[str, Union[Callable[[], Any], float]]


class AnyFunction(Protocol):
    def __call__(self, arg: Any) -> Any:
        pass


class KwargFunction(Protocol):
    def __call__(self, *args, **kwargs: Any) -> Any:
        pass



"""Mesh output in the STL format."""
from typing import Dict, List

import numpy as np
from stl import Mesh, stl
from tunable import Tunable

from ..simulation.simulator import World
from . import (
    Output,
    OutputReproducibleFiles,
    check_overwrite,
    ensure_path_and_extension_and_number,
)


class MeshCellScaleFactor(Tunable):
    """Scale factor for mesh output"""

    default: float = 1.0


class MeshOutput(Output):
    """Mesh output in the STL format."""

    def __init__(self):
        super().__init__()

    def output(self, world: World, **kwargs) -> List[Dict[str, np.ndarray]]:
        meshes = []

        # for boundary in world.boundaries:
        #     pass

        scale_props = ('width', 'length')

        for cell in world.cells:
            backup = {}

            if MeshCellScaleFactor.value != 1.0:
                for sp in scale_props:
                    if hasattr(cell, sp):
                        value = getattr(cell, sp)
                        backup[sp] = value
                        setattr(cell, sp, value * MeshCellScaleFactor.value)

            vertices, triangles = cell.points3d_on_canvas()

            for k, v in backup.items():
                setattr(cell, k, v)

            meshes.append(dict(vertices=vertices, triangles=triangles))

        return meshes

    def write(
        self,
        world: World,
        file_name: str,
        overwrite: bool = False,
        output_count: int = 0,
        **kwargs
    ) -> None:
        meshes = self.output(world)

        stl_meshes = []

        for single_mesh in meshes:
            vertices, triangles = single_mesh['vertices'], single_mesh['triangles']
            data = Mesh(np.zeros(len(triangles), dtype=Mesh.dtype))
            for i, tri in enumerate(triangles):
                data.vectors[i, :] = vertices[tri, :]

            stl_meshes.append(data)

        result_mesh = Mesh(
            np.concatenate([single_mesh.data for single_mesh in stl_meshes])
            if stl_meshes
            else np.zeros(0, dtype=Mesh.dtype)
        )

        if OutputReproducibleFiles.value:
            stl.HEADER_FORMAT = '{package_name} ({version})'

        result_mesh.save(
            check_overwrite(
                ensure_path_and_extension_and_number(file_name, '.stl', output_count),
                overwrite=overwrite,
            )
        )

    def display(self, world: World, **kwargs) -> None:
        # ax = fig.add_subplot(111, projection='3d')
        # ax.scatter(vertices[:, 0], vertices[:, 1], vertices[:, 2])
        # ax.set_aspect('equal', 'datalim')
        raise RuntimeError("Unsupported")


__all__ = ['MeshOutput']



"""Simulator base classes."""

import numpy as np

from ..parameters import s_to_h
from . import BaseSimulator


class World:
    """The World class contains the cells and, if present, the boundaries."""

    def __init__(self):
        self.cells = []
        self.boundaries = []

        self.cells_to_add = []
        self.cells_to_remove = []

    def add_boundary(self, coordinates: np.ndarray) -> None:
        """
        Add a boundary to the simulation.

        :param coordinates: Coordinates of the boundary.
        :return: None
        """
        self.boundaries.append(np.array(coordinates))

    def clear(self) -> None:
        """
        Resets the World.

        :return: None
        """
        self.cells.clear()
        self.cells_to_add.clear()
        self.cells_to_remove.clear()
        self.boundaries.clear()

    def add(self, cell: object) -> None:
        """
        Adds a cell to the World.

        :param cell: Cell
        :return: None
        """
        self.cells_to_add.append(cell)

    def remove(self, cell: object) -> None:
        """
        Removes a cell from the world.

        :param cell: Cell
        :return: None
        """
        self.cells_to_remove.append(cell)

    def commit(self) -> None:
        """
        Commits a step. Cells to be added, and cells to be removed,
        will only be applied once commit is called.

        :return: None
        """
        for cell in self.cells_to_add:
            self.cells.append(cell)
        for cell in self.cells_to_remove:
            self.cells.remove(cell)

        self.cells_to_remove.clear()
        self.cells_to_add.clear()

    def copy(self) -> "World":
        """
        Creates a copy of thw World.

        :return: Copy of the World
        """
        new_world = self.__class__()
        new_world.cells = self.cells[:]
        new_world.boundaries = self.boundaries[:]

        new_world.cells_to_add = self.cells_to_add[:]
        new_world.cells_to_remove = self.cells_to_remove[:]

        return new_world


class Simulation:
    """Simulation represents the simulation state at a certain timepoint,
    i.e. a World and a time."""

    def __init__(self):
        self.world = World()
        self.time = 0.0


class Timestep:
    """Timestep is an auxiliary class combining a certain
    timepoint, simulation and simulator."""

    __slots__ = "timestep", "simulation", "simulator"

    @property
    def hours(self) -> float:
        """
        The hours passed within this timestep.

        :return: Hours
        """
        return s_to_h(self.timestep)

    @property
    def time(self) -> float:
        """
        Total simulation time passed in seconds.

        :return: Seconds
        """
        return self.simulation.time

    @property
    def time_hours(self) -> float:
        """
        Total simulation time passed in hours.

        :return: Hours
        """
        return s_to_h(self.time)

    @property
    def world(self) -> World:
        return self.simulation.world

    def __init__(self, timestep: float, simulation: Simulation, simulator: "Simulator"):
        self.timestep, self.simulation, self.simulator = timestep, simulation, simulator


class Simulator(BaseSimulator):
    """Simulator class, a class serving as interface to World and sub-simulators
    (such as physical placement), as well as the caller of each cells step function."""

    def __init__(self):
        simulation = Simulation()
        self.simulation = simulation

        self.sub_simulators = []

    def add(self, cell: object) -> None:

        self.simulation.world.add(cell)

    def remove(self, cell: object) -> None:

        self.simulation.world.remove(cell)

    def add_boundary(self, coordinates: np.ndarray) -> None:

        self.simulation.world.add_boundary(np.array(coordinates))

    def clear(self) -> None:

        self.simulation.world.clear()

    def step(self, timestep: float = 0.0) -> Timestep:

        simulation = self.simulation
        # possibly a deep copy if former states should be preserved?

        simulation.time += timestep

        ts = Timestep(timestep, simulation, self)

        simulation.world.commit()

        for cell in simulation.world.cells:
            cell.step(ts)

        simulation.world.commit()

        for sim in self.sub_simulators:
            sim.clear()

            for boundary in self.simulation.world.boundaries:
                sim.add_boundary(boundary)

            for cell in simulation.world.cells:
                sim.add(cell)

            sim.step(timestep)

        return ts



from pathlib import Path

import cv2
import matplotlib.pyplot
import numpy as np
import pytest
from numpy.testing import assert_almost_equal

from ..output import (
    OutputReproducibleFiles,
    check_overwrite,
    ensure_extension,
    ensure_path,
)
from ..output.all import (
    COCOOutput,
    FluorescenceRenderer,
    GenericMaskOutput,
    JsonPickleSerializer,
    MeshOutput,
    NoisyUnevenIlluminationPhaseContrast,
    Output,
    PlainRenderer,
    QuickAndDirtyTableDumper,
    SvgRenderer,
    TiffOutput,
    TrackMateXML,
    YOLOOutput,
)
from ..output.gt import (
    COCOEncodeRLE,
    COCOOutputStuff,
    GroundTruthOnlyCompleteCellsInImages,
    GroundTruthOutput,
    binary_to_rle,
)
from ..output.mesh import MeshCellScaleFactor
from ..output.plot import PlotRenderer
from ..output.render import (
    OpenCVimshow,
    RenderChannels,
    RoiOutputScaleDelta,
    RoiOutputScaleFactor,
    add_if_uneven,
    bytescale,
    cv2_has_write_support,
    get_canvas_points_for_cell,
    get_canvas_points_raw,
    new_canvas,
    render_on_canvas_matplotlib,
    scale_points_absolute,
    scale_points_relative,
)
from ..output.serialization import type2numpy
from ..output.xml import TrackMateXMLExportFluorescences, TrackMateXMLExportLengthTypo


def test_jsonpickle(simulator, capsys):

    output = JsonPickleSerializer()
    output.display(simulator.simulation.world)

    captured = capsys.readouterr()

    result = captured.out.replace('\n', '')
    expected = (
        '{"py/object": "cellsium.simulation.simulator.World", "cells": [{"py/object":'
        ' "cellsium.cli.SizerCell", "lineage_history": [0], "id_": 1, "parent_id": 0, '
        '"birth_time": 0.0, "angle": 3.6942842669518385, "position": '
        '[17.90727229572483, 29.944407189601154], "bend_overall": 0.01753978255362451,'
        ' "bend_upper": -0.04126808565523994, "bend_lower": -0.09614075306364522, '
        '"length": 1.636246698756115, "width": 0.9548930351446809, "division_size": '
        '2.898390206023354, "elongation_rate": 1.5}], "boundaries": [], '
        '"cells_to_add": [], "cells_to_remove": []}'
    )

    assert result == expected


def test_type2numpy_nomaxlen():
    assert type2numpy([1, 2]) == '(2,)i8'


def test_type2numpy_nonsense():
    with pytest.raises(RuntimeError):
        type2numpy(type2numpy)


def test_qadtd_empty(simulator):
    for cell in simulator.simulation.world.cells:
        simulator.simulation.world.remove(cell)

    simulator.simulation.world.commit()

    output = QuickAndDirtyTableDumper()

    assert len(output.output(simulator.simulation.world)) == 0


def test_svgrenderer_display(simulator):
    output = SvgRenderer()

    with pytest.raises(RuntimeError):
        output.display(simulator.simulation.world)


def test_trackmate_display(simulator):
    output = TrackMateXML()

    with pytest.raises(RuntimeError):
        output.display(simulator.simulation.world)


def test_trackmate_no_fluor(simulator, tunables, tmpdir):
    testfile = tmpdir.join('testfile.xml')
    for _ in range(2):
        simulator.step(60.0 * 60.0)

    with tunables((TrackMateXMLExportFluorescences, '')):
        output = TrackMateXML()

        output.write(simulator.simulation.world, testfile)


def test_trackmate_typo(simulator, tunables, tmpdir):
    testfile = tmpdir.join('testfile.xml')
    for _ in range(2):
        simulator.step(60.0 * 60.0)

    with tunables((TrackMateXMLExportLengthTypo, False)):
        output = TrackMateXML()

        output.write(simulator.simulation.world, testfile)


def test_trackmate_growth(simulator, tunables, tmpdir):
    testfile = tmpdir.join('testfile.xml')

    output = TrackMateXML()

    for _ in range(5):
        simulator.step(60.0 * 60.0)
        output.write(simulator.simulation.world, testfile, overwrite=True)


def test_gt_output(simulator):
    output = GroundTruthOutput()

    with pytest.raises(RuntimeError):
        output.output(simulator.simulation.world)


def test_gt_virtual():
    output = GroundTruthOutput()

    output._write_initializations(None, '')
    output._write_perform(None, '')


def test_yolo_off_canvas_cells(simulator, tmpdir, tunables):
    testdir = tmpdir.join('yoloout')

    output = YOLOOutput()

    simulator.step(60.0 * 60.0)

    # move one cell off canvas
    some_cell_position = simulator.simulation.world.cells[1].position
    some_cell_position[0] += 10000

    output.write(simulator.simulation.world, testdir)

    with tunables((GroundTruthOnlyCompleteCellsInImages, False)):
        output.write(simulator.simulation.world, testdir, overwrite=True)


def test_yolo_no_overwrite(simulator, tmpdir, tunables):
    with tunables((RenderChannels, 'PlainRenderer')):
        testdir = tmpdir.join('yoloout')

        output = YOLOOutput()

        output.write(simulator.simulation.world, testdir)

        output = YOLOOutput()

        with pytest.raises(RuntimeError):
            output.write(simulator.simulation.world, testdir)


def test_coco_no_overwrite(simulator, tmpdir, tunables):
    with tunables((RenderChannels, 'PlainRenderer')):

        testdir = tmpdir.join('cocoout')

        output = COCOOutput()
        output.write(simulator.simulation.world, testdir)

        output = COCOOutput()

        with pytest.raises(RuntimeError):
            output.write(simulator.simulation.world, testdir)


def test_coco_unreproducible_stuffthings(simulator, tmpdir, tunables):
    with tunables(
        (RenderChannels, 'PlainRenderer'),
        (OutputReproducibleFiles, False),
        (COCOOutputStuff, True),
        (COCOEncodeRLE, True),
    ):
        testdir = tmpdir.join('cocoout')

        output = COCOOutput()

        simulator.step(60.0 * 60.0)

        # move one cell off canvas
        some_cell_position = simulator.simulation.world.cells[1].position
        some_cell_position[0] += 10000

        output.write(simulator.simulation.world, testdir)

        with tunables((GroundTruthOnlyCompleteCellsInImages, False)):
            output.write(simulator.simulation.world, testdir, overwrite=True)


def test_genericmaskoutput_no_overwrite(simulator, tmpdir, tunables):
    with tunables((RenderChannels, 'PlainRenderer')):

        testdir = tmpdir.join('cocoout')

        output = GenericMaskOutput()

        output.write(simulator.simulation.world, testdir)

        output = GenericMaskOutput()

        with pytest.raises(RuntimeError):
            output.write(simulator.simulation.world, testdir)


def test_mesh_scale(reset_state, simulator, tmpdir, tunables):
    testfile = tmpdir.join('testfile.stl')

    with tunables((MeshCellScaleFactor, 1.1)):

        output = MeshOutput()

        output.write(simulator.simulation.world, file_name=str(testfile))


def test_mesh_output():
    output = MeshOutput()

    with pytest.raises(RuntimeError):
        output.display(None)


def test_plotrenderer_quit(simulator, monkeypatch, nop):
    monkeypatch.setattr(matplotlib.pyplot, 'ion', nop)
    monkeypatch.setattr(matplotlib.pyplot, 'show', nop)

    output = PlotRenderer()

    output.display(simulator.simulation.world)

    matplotlib.pyplot.close(output.fig.number)

    with pytest.raises(KeyboardInterrupt):
        output.display(simulator.simulation.world)


def test_mesh_cell_zoo(reset_state, simulator, tmpdir, add_cell_zoo):
    testfile = tmpdir.join('testfile.stl')

    add_cell_zoo(simulator)

    output = MeshOutput()

    output.write(simulator.simulation.world, file_name=str(testfile))


def test_render_debug(reset_state, simulator, tmpdir, add_cell_zoo, chdir):
    new_pwd = tmpdir.mkdir('debugout')
    testfile = tmpdir.join('testfile.png')

    add_cell_zoo(simulator)

    output = NoisyUnevenIlluminationPhaseContrast()
    output.write_debug_output = True

    with chdir(str(new_pwd)):
        output.write(simulator.simulation.world, file_name=str(testfile))

    assert len(new_pwd.listdir()) > 0


def test_render_bytescale_nochange():
    input_data = np.zeros((256, 256), dtype=np.uint8)

    assert id(bytescale(input_data)) == id(input_data)


def test_render_add_if_uneven():
    assert add_if_uneven(1) == 2
    assert add_if_uneven(2) == 2


def test_fluorescence_not_the_right_channel(
    reset_state, simulator, tmpdir, add_cell_zoo
):
    testfile = tmpdir.join('testfile.png')

    add_cell_zoo(simulator)

    output = FluorescenceRenderer()
    output.channel = 1

    output.write(simulator.simulation.world, testfile)


def test_render_multichannel_tif(
    reset_state, simulator, tmpdir, add_cell_zoo, tunables
):
    testfile = tmpdir.join('testfile.tif')

    add_cell_zoo(simulator)

    with tunables(
        (RenderChannels, 'NoisyUnevenIlluminationPhaseContrast, FluorescenceRenderer')
    ):
        output = TiffOutput()

        output.write(simulator.simulation.world, file_name=str(testfile))


def test_render_renderchannels_nonsense(tunables):
    assert not RenderChannels.test('foo')


def test_render_mock_failcv2(monkeypatch):
    def raise_cv2_error(*args, **kwargs):
        raise cv2.error()

    monkeypatch.setattr(cv2, 'haveImageWriter', raise_cv2_error)

    assert not cv2_has_write_support('.foo')


def test_render_scale_points_relative():
    points = np.linspace(0, 1)
    points = np.c_[points, points]
    scale_factor = 1.1
    scaled_points = scale_points_relative(points, scale_factor)

    assert_almost_equal(scaled_points.min(), -0.05)
    assert_almost_equal(scaled_points.max(), 1.05)


def test_render_scale_points_absolute_nop():
    points = np.linspace(0, 1)
    points = np.c_[points, points]

    scaled_points = scale_points_absolute(points)

    assert id(scaled_points) == id(points)


def test_render_get_canvas_points_for_cell(simulator, tunables):
    cell = simulator.simulation.world.cells[0]

    with tunables((RoiOutputScaleFactor, 1.0)):
        get_canvas_points_for_cell(cell)

    with tunables((RoiOutputScaleFactor, 1.1)):
        get_canvas_points_for_cell(cell)

    with tunables((RoiOutputScaleDelta, 0.0)):
        get_canvas_points_for_cell(cell)

    with tunables((RoiOutputScaleDelta, 0.1)):
        get_canvas_points_for_cell(cell)


def test_render_plainoutput_display_cv2(simulator, tunables, monkeypatch, nop):

    output = PlainRenderer()

    with tunables((OpenCVimshow, True)):
        monkeypatch.setattr(cv2, 'imshow', nop)
        monkeypatch.setattr(cv2, 'waitKey', nop)

        output.display(simulator.simulation.world)


def test_render_plainoutput_display_matplotlib(simulator, tunables, monkeypatch, nop):

    output = PlainRenderer()

    with tunables((OpenCVimshow, False)):
        monkeypatch.setattr(matplotlib.pyplot, 'ion', nop)
        monkeypatch.setattr(matplotlib.pyplot, 'show', nop)

        output.display(simulator.simulation.world)

        # the second call will be handled differently, hence we try it as well

        output.display(simulator.simulation.world)


def test_render_on_canvas_matplotlib_obscure_options(simulator, monkeypatch, nop):
    canvas = new_canvas()
    array_of_points = [
        get_canvas_points_raw(cell, canvas.shape[0])
        for cell in simulator.simulation.world.cells
    ]

    monkeypatch.setattr(matplotlib.pyplot, 'ion', nop)
    monkeypatch.setattr(matplotlib.pyplot, 'ioff', nop)

    monkeypatch.setattr(matplotlib.pyplot, 'isinteractive', lambda: True)

    render_on_canvas_matplotlib(canvas, array_of_points, over_sample=2)


def test_rle():
    mask = np.zeros((128, 128), dtype=bool)
    mask[32 : 32 + 64, 32 : 32 + 64] = 1

    result = binary_to_rle(mask)

    assert result.ravel().tolist() == ([4128] + ([64] * 127) + [4128])


def test_output_dummy():
    output = object.__new__(Output)
    output.output(None)
    output.write(None, None)

    with pytest.raises(RuntimeError):
        output.display(None)


def test_overwrite(tmpdir):
    testfile = tmpdir.join('testfile')
    testfile.write("Test")

    with pytest.raises(RuntimeError):
        check_overwrite(str(testfile), overwrite=False)


def test_ensure_extension():
    assert ensure_extension('Hello{}.txt', '.txt') == 'Hello.txt'


def test_mkdir(tmpdir):
    p = Path(str(tmpdir)) / "some" / "directories"

    some_file = p / "some_file"

    assert not p.exists()

    ensure_path(str(some_file))

    assert p.exists()



import pytest

from ..cli.cli import main
from ..output import Output
from ..simulation.placement.base import PlacementSimulationSimplification

ALL_OUTPUTS = [
    'COCOOutput',
    'FluorescenceRenderer',
    'GenericMaskOutput',
    'JsonPickleSerializer',
    'CsvOutput',
    'MeshOutput',
    'NoisyUnevenIlluminationPhaseContrast',
    'PhaseContrastRenderer',
    'PlainRenderer',
    'PlotRenderer',
    'QuickAndDirtyTableDumper',
    'SvgRenderer',
    'TiffOutput',
    'TrackMateXML',
    'UnevenIlluminationPhaseContrast',
    'YOLOOutput',
]


def _generate_tunables(t):
    args = []
    if t:
        for k, v in t.items():
            args.append('-t')
            args.append('%s=%s' % (k, v))
    return args


def _generate_output(Output):
    args = []
    if Output:
        for an_output in Output:
            args.append('--Output')
            args.append(an_output)
    return args


def generate_commandline(
    task='training',
    prefix=False,
    overwrite=False,
    verbose=False,
    quiet=False,
    t=None,
    output=None,
    Output=None,
    PlacementSimulation=None,
    args=None,
):
    if args is None:
        additional_args = []
    else:
        additional_args = args

    args = [task]
    if prefix:
        args.append('--prefix')

    if overwrite:
        args.append('--overwrite')

    if verbose:
        args.append('--verbose')

    if quiet:
        args.append('--quiet')

    args += _generate_tunables(t)

    if output:
        args.append('--output')
        args.append(output)

    args += _generate_output(Output)

    if PlacementSimulation:
        args.append('--PlacementSimulation')
        args.append(PlacementSimulation)

    args += additional_args

    return args


def call_main(*args, **kwargs):
    args = generate_commandline(*args, **kwargs)

    return_value = main(args)

    assert return_value is None or return_value == 0


def test_training_outputs(reset_state, tmpdir):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'training',
        prefix=True,
        overwrite=True,
        t=dict(TrainingDataCount=2, TrainingImageWidth=64, TrainingImageHeight=64),
        output=str(output_dir) + '/output_name',
        Output=ALL_OUTPUTS,
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0


def test_quiet_switch(reset_state, tmpdir):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'training',
        quiet=True,
        t=dict(TrainingDataCount=2),
        output=str(output_dir) + '/output_name',
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0


def test_verbose_switch(reset_state, tmpdir):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'training',
        verbose=True,
        t=dict(TrainingDataCount=2),
        output=str(output_dir) + '/output_name',
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0


def test_debug_switch(reset_state, tmpdir):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'training',
        t=dict(TrainingDataCount=2),
        output=str(output_dir) + '/output_name',
        args=['-vv'],
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0


def test_training_no_output_set(reset_state):
    with pytest.raises(RuntimeError):
        call_main('training', t=dict(TrainingDataCount=2), Output=ALL_OUTPUTS)


def test_simulation_output_no_prefix(reset_state, tmpdir):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'simulate',
        overwrite=True,
        t=dict(
            SimulationTimestep=0.1, SimulationOutputInterval=0.1, SimulationDuration=0.2
        ),
        output=str(output_dir) + '/output_name',
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0


def test_simulation_outputs(reset_state, tmpdir):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'simulate',
        prefix=True,
        overwrite=True,
        t=dict(
            SimulationTimestep=0.1,
            SimulationOutputInterval=0.1,
            SimulationDuration=0.2,
            SimulationOutputFirstState=1,
            Width=10,
            Height=10,
        ),
        output=str(output_dir) + '/output_name',
        Output=ALL_OUTPUTS,
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0


@pytest.mark.parametrize('s', [2, 1])
def test_simulation_placementsimplification(s, reset_state, tmpdir):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'simulate',
        prefix=True,
        overwrite=True,
        t=dict(
            SimulationTimestep=0.1,
            SimulationOutputInterval=0.1,
            SimulationDuration=0.2,
            PlacementSimulationSimplification=s,
        ),
        output=str(output_dir) + '/output_name',
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0


class InterruptOutput(Output):
    interrupt_at = 0

    def __init__(self):
        self.round = 0

    #     def display(self, world, **kwargs):
    #         self.output(world, **kwargs)

    def write(self, world, file_name, **kwargs):
        self.output(world, **kwargs)

    def output(self, world, **kwargs):
        if self.__class__.interrupt_at == self.round:
            raise KeyboardInterrupt
        self.round += 1


@pytest.mark.parametrize('interrupt_at', [0, 1])
def test_simulation_keyboard_interrupt(interrupt_at, reset_state, tmpdir):
    # this test is somewhat slow, but quite important to check whether all
    # outputs handle deallocation properly
    # (either before or after they had something to do)
    output_dir = tmpdir.mkdir('result')

    InterruptOutput.interrupt_at = interrupt_at

    call_main(
        'simulate',
        prefix=True,
        overwrite=True,
        t=dict(
            SimulationTimestep=0.1,
            SimulationOutputInterval=0.1,
            SimulationDuration=-1,
            Width=10,
            Height=10,
        ),
        output=str(output_dir) + '/output_name',
        Output=['InterruptOutput'] + ALL_OUTPUTS,
    )

    generated_files = output_dir.listdir()

    print(generated_files)

    if interrupt_at == 0:
        assert len(generated_files) == 0
    else:
        assert len(generated_files) > 0


@pytest.mark.parametrize('dxftype', ['LWPOLYLINE'])
def test_simulation_outputs_with_boundary(dxftype, reset_state, tmpdir, dxf_file):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'simulate',
        prefix=True,
        overwrite=True,
        t=dict(
            SimulationTimestep=0.1,
            SimulationOutputInterval=0.1,
            SimulationDuration=0.2,
            BoundariesFile=str(dxf_file(dxftype)),
            Width=20,
            Height=20,
        ),
        output=str(output_dir) + '/output_name',
        Output=ALL_OUTPUTS,
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0


@pytest.mark.parametrize('dxftype', ['LWPOLYLINE', 'POLYLINE', 'SPLINE'])
def test_simulation_with_boundary(dxftype, reset_state, tmpdir, dxf_file):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'simulate',
        prefix=True,
        overwrite=True,
        t=dict(
            SimulationTimestep=0.1,
            SimulationOutputInterval=0.1,
            SimulationDuration=0.2,
            BoundariesFile=str(dxf_file(dxftype)),
        ),
        output=str(output_dir) + '/output_name',
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0


@pytest.mark.parametrize('dxftype', ['LWPOLYLINE'])
def test_simulation_with_boundary_box2d(dxftype, reset_state, tmpdir, dxf_file):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'simulate',
        prefix=True,
        overwrite=True,
        t=dict(
            SimulationTimestep=0.1,
            SimulationOutputInterval=0.1,
            SimulationDuration=0.2,
            BoundariesFile=str(dxf_file(dxftype)),
            PlacementSimulationSimplification=2,
        ),
        PlacementSimulation='Box2D',
        output=str(output_dir) + '/output_name',
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0


def test_render(reset_state, tmpdir):
    output_dir = tmpdir.mkdir('result')

    call_main(
        'training',
        overwrite=True,
        t=dict(TrainingDataCount=1),
        output=str(output_dir) + '/output_name',
        Output=['JsonPickleSerializer'],
    )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0

    reset_state()

    output_dir_2 = tmpdir.mkdir('result2')

    call_main(
        'render',
        output=str(output_dir_2) + '/output_name',
        Output=['GenericMaskOutput'],
        args=['--input-file', str(output_dir) + '/output_name000.json'],
    )

    generated_files = output_dir_2.listdir()

    assert len(generated_files) > 0


@pytest.mark.parametrize('s', [2, 1])
def test_training_box2d_placementsimplification(s, reset_state, tmpdir, tunables):
    output_dir = tmpdir.mkdir('result')

    with tunables((PlacementSimulationSimplification, s)):
        call_main(
            'simulate',
            prefix=True,
            overwrite=True,
            PlacementSimulation='Box2D',
            t=dict(
                SimulationTimestep=0.1,
                SimulationOutputInterval=0.1,
                SimulationDuration=0.2,
                PlacementSimulationSimplification=s,
            ),
            output=str(output_dir) + '/output_name',
        )

    generated_files = output_dir.listdir()

    assert len(generated_files) > 0



def generate_dxf_file(output_filename, line_type='LWPOLYLINE'):
    assert line_type in ('LWPOLYLINE', 'POLYLINE', 'SPLINE')

    x, y = 10, 10
    w, h = 20, 20

    lines = [
        [[x, y], [y, y + h]],
        [[x, y + h], [x + w, y + h]],
        [[x + w, y + h], [x + w, y]],
        [[x + w, y], [x, y]],
    ]

    import ezdxf

    dxf = ezdxf.new('R2010')

    modelspace = dxf.modelspace()

    if line_type == 'LWPOLYLINE':
        for line_points in lines:
            modelspace.add_lwpolyline(line_points)
    elif line_type == 'POLYLINE':
        for line_points in lines:
            modelspace.add_polyline2d(line_points)
    elif line_type == 'SPLINE':
        for line_points in lines:
            modelspace.add_spline(line_points)
    dxf.saveas(output_filename)


if __name__ == '__main__':
    import sys

    if len(sys.argv) == 1:
        print(f"Usage: {sys.argv[0]} filename.dxf <LWPOLYLINE|POLYLINE|SPLINE>")
    elif len(sys.argv) == 2:
        generate_dxf_file(sys.argv[1])
    elif len(sys.argv) > 2:
        generate_dxf_file(sys.argv[1], sys.argv[2])



"""Cell model package."""
from typing import Any, Iterable, Mapping, Optional

from ..parameters import h_to_s, s_to_h
from ..simulation.simulator import Timestep
from .agent import (
    Copyable,
    IdCounter,
    InitializeWithParameters,
    Representable,
    WithLineage,
    WithLineageHistory,
    WithRandomSequences,
    WithTemporalLineage,
)
from .geometry import (
    AutoMesh3D,
    BentRod,
    CellGeometry,
    Coccoid,
    Ellipsoid,
    Rectangle,
    RodShaped,
    Shape,
    Shape3D,
    Square,
    WithAngle,
    WithFluorescence,
    WithPosition,
    WithProperDivisionBehavior,
)
from .initialization import (
    RandomAngle,
    RandomBentRod,
    RandomPosition,
    RandomWidthLength,
)


def generate_cell(*additional_classes: type, name: str = "PlacedCell"):
    """
    Generates a cell class using the standard classes, and possible additional classes.

    :param additional_classes: Additional classes to inherit the cell from.
    :param name: Name of the class
    :return: Class
    """
    cell = type(
        name,
        (
            WithLineageHistory,
            WithLineage,
            WithTemporalLineage,
            WithProperDivisionBehavior,
            InitializeWithParameters,
            Copyable,
            Representable,
            WithRandomSequences,
            RandomWidthLength,
            RandomBentRod,
            RandomPosition,
            RandomAngle,
            CellGeometry,
        )
        + additional_classes,
        {},
    )

    return cell


PlacedCell = generate_cell(BentRod)


def assemble_cell(
    simulated_model: type,
    *additional_classes,
    placed_cell: type = PlacedCell,
    name: str = "Cell"
):
    """
    Assembles a cell class from parent classes.
    Necessary as the cell class needs the right level of inheritance.

    :param simulated_model: Model class
    :param additional_classes: Classes to create a cell type, or
    :param placed_cell: A prepared cell
    :param name: Optional name
    :return: Cell class
    """

    if additional_classes:
        base_cell = generate_cell(*additional_classes)
    else:
        base_cell = placed_cell

    return type(name, (simulated_model, base_cell), {})


class SimulatedCell:
    """
    Base class for simulated cells, allowing for division behavior.
    """

    def birth(
        self, parent: Optional["SimulatedCell"] = None, ts: Optional[Timestep] = None
    ) -> None:
        """
        Called when a cell is "born".

        :param parent: Parent cell
        :param ts: Timestep
        :return: None
        """
        pass

    def grow(self, ts: Timestep) -> None:
        """
        Called each timestep to grow cell.

        :param ts: Timestep
        :return: None
        """
        pass

    def divide(self, ts: Timestep) -> Iterable["SimulatedCell"]:
        """
        Called when a cell should divide, creates the daughter cells.

        :param ts: Timestep
        :return: None
        """
        offspring_a, offspring_b = self.copy(), self.copy()

        offspring_a.position, offspring_b.position = self.get_division_positions()

        if isinstance(self, WithLineage):
            offspring_a.parent_id = offspring_b.parent_id = self.id_

        if isinstance(self, WithLineageHistory):
            offspring_a.lineage_history = self.lineage_history[:] + [self.id_]
            offspring_b.lineage_history = self.lineage_history[:] + [self.id_]

        if isinstance(self, WithTemporalLineage):
            now = ts.simulation.time
            offspring_b.birth_time = offspring_a.birth_time = now

        ts.simulator.add(offspring_a)
        ts.simulator.add(offspring_b)

        offspring_a.birth(parent=self, ts=ts)
        offspring_b.birth(parent=self, ts=ts)

        ts.simulator.remove(self)

        return offspring_a, offspring_b

    def step(self, ts: Timestep) -> None:
        """
        Timestep function of the cell object, called by the simulator.

        :param ts: Timestep
        :return: None
        """
        self.grow(ts=ts)


# noinspection PyAttributeOutsideInit
class SizerCell(SimulatedCell):
    """
    Example cell implementing a simple sizer growth mechanism.
    """

    @staticmethod
    def random_sequences(sequence: Any) -> Mapping[str, Any]:
        return dict(division_size=sequence.normal(3.0, 0.25))  # µm

    def birth(
        self, parent: Optional["SizerCell"] = None, ts: Optional[Timestep] = None
    ) -> None:
        self.division_size = next(self.random.division_size)
        self.elongation_rate = 1.5

    def grow(self, ts: Timestep) -> None:
        self.length += self.elongation_rate * ts.hours

        if self.length > self.division_size:
            offspring_a, offspring_b = self.divide(ts)
            offspring_a.length = offspring_b.length = self.length / 2


# noinspection PyAttributeOutsideInit
class TimerCell(SimulatedCell):
    """
    Example cell implementing a simple timer growth mechanism.
    """

    @staticmethod
    def random_sequences(sequence: Any) -> Mapping[str, Any]:
        return dict(elongation_rate=sequence.normal(1.5, 0.25))  # µm·h⁻¹

    def birth(
        self, parent: Optional["TimerCell"] = None, ts: Optional[Timestep] = None
    ) -> None:
        self.elongation_rate = next(self.random.elongation_rate)
        self.division_time = h_to_s(1.0)

    def grow(self, ts: Timestep) -> None:
        self.length += self.elongation_rate * ts.hours

        if ts.time > (self.birth_time + self.division_time):
            offspring_a, offspring_b = self.divide(ts)
            offspring_a.length = offspring_b.length = self.length / 2


__all__ = [
    'InitializeWithParameters',
    'WithRandomSequences',
    'Copyable',
    'Representable',
    'IdCounter',
    'WithLineage',
    'WithLineageHistory',
    'WithTemporalLineage',
    'Shape',
    'Shape3D',
    'RodShaped',
    'Rectangle',
    'Square',
    'BentRod',
    'Coccoid',
    'Ellipsoid',
    'WithPosition',
    'WithAngle',
    'WithFluorescence',
    'WithProperDivisionBehavior',
    'AutoMesh3D',
    'CellGeometry',
    's_to_h',
    'h_to_s',
    'PlacedCell',
    'SimulatedCell',
    'TimerCell',
    'SizerCell',
    'generate_cell',
    'assemble_cell',
    'Timestep',
]



"""Rendering CLI utility, render a simulation state saved using jsonpickle."""

from argparse import ArgumentParser, Namespace

import jsonpickle

from ...output import Output


def subcommand_argparser(parser: ArgumentParser) -> None:
    """
    Handle the argument parser for the 'render' subcommand.

    :param parser: Argument parser
    :return: None
    """
    parser.add_argument('-i', '--input-file', dest='input', default=None, required=True)


def subcommand_main(args: Namespace) -> None:
    """
    Entry point for the 'render' subcommand.

    :param args: Pre-parsed arguments
    :return: None
    """
    output = Output()

    with open(args.input, 'r') as fp:
        data = fp.read()

    data = jsonpickle.decode(data)

    output.write(data, args.output)



import sys
import types

import cellsium

from . import collect_modules_recursive, run_tests_recursively


def test_doctest():
    # chain run
    from . import run_doctests

    assert run_doctests(quiet=True) == 0


def test_doctest_cmr_str():
    result = collect_modules_recursive(cellsium, blacklist='cellsium')
    assert result == []


def test_doctest_cmr_list():
    result = collect_modules_recursive(cellsium, blacklist=['cellsium'])
    assert result == []


def test_doctest_fail_a_test(patch_sys_modules):
    def a_test_function():
        """
        Fail a doctest.

        >>> a_test_function()
        43
        """
        return 42

    mock_module = types.ModuleType('MockModule', 'Just a mock module')
    mock_module.__dict__['a_test_function'] = a_test_function
    a_test_function.__module__ = mock_module.__name__

    with patch_sys_modules():
        sys.modules[mock_module.__name__] = mock_module
        assert run_tests_recursively(mock_module) == 1



"""Cell parameter random initializations."""
from typing import Any, Dict

import numpy as np

from ..parameters import (
    Height,
    NewCellBendLowerLower,
    NewCellBendLowerUpper,
    NewCellBendOverallLower,
    NewCellBendOverallUpper,
    NewCellBendUpperLower,
    NewCellBendUpperUpper,
    NewCellLength1Mean,
    NewCellLength1Std,
    NewCellLength2Mean,
    NewCellLength2Std,
    NewCellLengthAbsoluteMax,
    NewCellLengthAbsoluteMin,
    NewCellRadiusFromCenter,
    NewCellWidthAbsoluteMax,
    NewCellWidthAbsoluteMin,
    NewCellWidthMean,
    NewCellWidthStd,
    Width,
)
from ..random import RRF, enforce_bounds

RandomSequenceType = Dict[str, Any]


class RandomWidthLength:
    """Random initializations for cell width/lengths."""

    @staticmethod
    def random_sequences(sequence: RRF) -> RandomSequenceType:

        assert NewCellLength1Mean.value > NewCellWidthMean.value
        assert NewCellLength2Mean.value > NewCellWidthMean.value

        def ensure_length_greater_width(length, width):
            for inner_length, inner_width in zip(length, width):
                if inner_length > inner_width:
                    yield [inner_length, inner_width]

        return dict(
            length__width=RRF.chain(
                ensure_length_greater_width,
                length=RRF.compose(
                    lambda raw_lengths, choice: raw_lengths[choice],
                    raw_lengths=RRF.chain(
                        enforce_bounds,
                        iterator=sequence.multivariate_normal(
                            [NewCellLength1Mean.value, NewCellLength2Mean.value],
                            [
                                [NewCellLength1Std.value, 0.0],
                                [0.0, NewCellLength2Std.value],
                            ],
                        ),
                        minimum=NewCellLengthAbsoluteMin.value,
                        maximum=NewCellLengthAbsoluteMax.value,
                    ),
                    choice=sequence.integers(0, 1),
                ),
                width=RRF.chain(
                    enforce_bounds,
                    iterator=sequence.normal(
                        NewCellWidthMean.value, NewCellWidthStd.value
                    ),
                    minimum=NewCellWidthAbsoluteMin.value,
                    maximum=NewCellWidthAbsoluteMax.value,
                ),
            )
        )


class RandomBentRod:
    """Random initializations for cell bent radii."""

    @staticmethod
    def random_sequences(sequence: RRF) -> RandomSequenceType:
        return dict(
            bend_overall=sequence.uniform(
                NewCellBendOverallLower.value,
                NewCellBendOverallUpper.value,
            ),
            bend_upper=sequence.uniform(
                NewCellBendUpperLower.value, NewCellBendUpperUpper.value
            ),
            bend_lower=sequence.uniform(
                NewCellBendLowerLower.value, NewCellBendLowerUpper.value
            ),
        )


class RandomPosition:
    """Random initializations for cell positions."""

    @staticmethod
    def random_sequences(sequence: RRF) -> RandomSequenceType:
        return dict(
            position=RRF.compose(
                lambda radius, angle: [
                    float(radius * np.cos(angle) + Width.value / 2),
                    float(radius * np.sin(angle) + Height.value / 2),
                ],
                radius=sequence.uniform(0, NewCellRadiusFromCenter.value),
                angle=RRF.wrap(sequence.uniform(0, 360.0), np.radians),
            )
        )


class RandomAngle:
    """Random initializations for cell angles."""

    @staticmethod
    def random_sequences(sequence: RRF) -> RandomSequenceType:
        return dict(angle=RRF.wrap(sequence.uniform(0, 360.0), np.radians))


class RandomFluorescence:
    """Random initializations for fluorescences."""

    @staticmethod
    def random_sequences(sequence: RRF) -> RandomSequenceType:
        return dict(fluorescences=sequence.uniform(0, 360.0, (1,)))



"""Main set of parameters for the simulation."""
from typing import Union

import numpy as np
from tunable import Tunable

NumericType = Union[float, np.ndarray]


class RandomlyDistributed(Tunable):
    """Parent class for randomly distributed tunables"""

    pass


class NewCellCount(Tunable):
    """New cells to add to the simulation"""

    default: int = 1


class NewCellRadiusFromCenter(RandomlyDistributed):
    """Maximum radius for new cells to be spawned from the origin"""

    default: float = 5.0


class NewCellWidthMean(RandomlyDistributed):
    """Mean cell width for new cells"""

    default: float = 1.0


class NewCellWidthStd(RandomlyDistributed):
    """Standard deviation of the width of new cells"""

    default: float = 0.1


class NewCellWidthAbsoluteMax(Tunable):
    """Absolute maximum width of new cells"""

    default: float = 1.5


class NewCellWidthAbsoluteMin(Tunable):
    """Absolute minimum width of new cells"""

    default: float = 0.75


class NewCellLength1Mean(RandomlyDistributed):
    """Mean cell length, subtype one"""

    default: float = 2.5


class NewCellLength1Std(RandomlyDistributed):
    """Standard deviation of the cell length, subtype one"""

    default: float = 0.15


class NewCellLength2Mean(RandomlyDistributed):
    """Mean cell length, subtype two"""

    default: float = 1.25


class NewCellLength2Std(RandomlyDistributed):
    """Standard deviation of the cell length, subtype one"""

    default: float = 0.15


class NewCellLengthAbsoluteMax(Tunable):
    """Absolute maximum length of new cells"""

    default: float = 3.5


class NewCellLengthAbsoluteMin(Tunable):
    """Absolute minimum length of new cells"""

    default: float = 0.8


class NewCellBendOverallLower(RandomlyDistributed):
    """Bend factor minimum for new new cells"""

    default: float = -0.1


class NewCellBendOverallUpper(RandomlyDistributed):
    """Bend factor maximum for new new cells"""

    default: float = 0.1


class NewCellBendUpperLower(RandomlyDistributed):
    """Bend factor minimum for the upper part of new new cells"""

    default: float = -0.1


class NewCellBendUpperUpper(RandomlyDistributed):
    """Bend factor maximum for the upper part of new new cells"""

    default: float = 0.1


class NewCellBendLowerLower(RandomlyDistributed):
    """Bend factor minimum for the lower part of new new cells"""

    default: float = -0.1


class NewCellBendLowerUpper(RandomlyDistributed):
    """Bend factor maximum for the lower part of new new cells"""

    default: float = 0.1


class Width(Tunable):
    """Width of the (outputted) simulation"""

    default: float = 40.0


class Height(Tunable):
    """Height of the (outputted) simulation"""

    default: float = 60.0


class Calibration(Tunable):
    """Calibration for outputs, micrometer per pixel"""

    default: float = 0.065


def pixel_to_um(pix: NumericType) -> NumericType:
    """
    Convert pixel to micrometer.

    :param pix: Pixel value
    :return: Micrometer value
    """
    return pix * Calibration.value


def um_to_pixel(um: NumericType) -> NumericType:
    """
    Convert micrometer to pixel.

    :param um: Micrometer value
    :return: Pixel value
    """
    return um / Calibration.value


def h_to_s(hours: NumericType) -> NumericType:
    """
    Convert hours to seconds.

    :param hours: Hours
    :return: Seconds
    """
    return hours * 60.0 * 60.0


def s_to_h(seconds: NumericType) -> NumericType:
    """
    Convert seconds to hours.

    :param seconds: Seconds
    :return: Hours
    """
    return seconds / (60.0 * 60.0)



"""Simulation package contains the simulation/simulator-related classes."""
import numpy as np


class BaseSimulator:
    def add(self, cell: object) -> None:
        """
        Add a cell to the simulation.

        :param cell: Cell
        :return: None
        """
        pass

    def remove(self, cell) -> None:
        """
        Remove a cell from the simulation.

        :param cell: Cell
        :return: None
        """
        pass

    def add_boundary(self, coordinates: np.ndarray) -> None:
        """
        Add a boundary to the simulation.

        :param coordinates: Coordinates of the boundary
        :return: None
        """
        pass

    def clear(self) -> None:
        """
        Clear the (world of the) simulation.

        :return: None
        """
        pass

    def step(self, timestep: float) -> None:
        """
        Advance the simulation by a timestep.

        :param timestep: Time passed in seconds
        :return: None
        """
        pass



"""Serialization outputs."""
import csv
from typing import Any, Dict, List, Mapping, Optional, Tuple

import jsonpickle
import jsonpickle.ext.numpy
import jsonpickle.util
import numpy as np

from ..simulation.simulator import World
from . import Output, check_overwrite, ensure_path_and_extension_and_number

jsonpickle.ext.numpy.register_handlers()
additional_primitives = (
    np.float32,
    np.float64,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
)

jsonpickle.util.PRIMITIVES = type(jsonpickle.util.PRIMITIVES)(
    list(jsonpickle.util.PRIMITIVES) + list(additional_primitives)
)


class JsonPickleSerializer(Output):
    """Output as jsonpickle serialized files."""

    def output(self, world: World, **kwargs) -> str:
        return jsonpickle.dumps(world)

    def write(
        self,
        world: World,
        file_name: str,
        overwrite: bool = False,
        output_count: int = 0,
        **kwargs
    ) -> None:
        with open(
            check_overwrite(
                ensure_path_and_extension_and_number(file_name, '.json', output_count),
                overwrite=overwrite,
            ),
            'w+',
        ) as fp:
            fp.write(self.output(world))

    def display(self, world: World, **kwargs) -> None:
        print(self.output(world))


def type2numpy(value: Any, max_len: int = None) -> str:
    if isinstance(value, int):
        return 'i8'
    elif isinstance(value, float):
        return 'f8'
    elif isinstance(value, list):
        if max_len is None:
            max_len = len(value)
        return '(' + str(max_len) + ',)' + type2numpy(value[0])
    else:
        raise RuntimeError('...')


def prepare_numpy_dtype(
    inner: Mapping[str, Any], list_max_lens: Optional[Mapping[str, int]] = None
) -> List[Tuple[str, str]]:
    return [
        (
            key,
            type2numpy(
                value,
                max_len=list_max_lens[key]
                if list_max_lens and key in list_max_lens
                else None,
            ),
        )
        for key, value in sorted(inner.items())
    ]


class QuickAndDirtyTableDumper(Output):
    """Simple tabular output."""

    def output(self, world: World, **kwargs) -> np.ndarray:
        if not world.cells:
            return np.zeros(0)

        first_cell = world.cells[0]
        list_lens = [
            {
                key: len(value)
                for key, value in celldict.items()
                if isinstance(value, list)
            }
            for celldict in (cell.__dict__ for cell in world.cells)
        ]

        list_max_lens = {
            key: max([list_len[key] for list_len in list_lens])
            for key in list_lens[0].keys()
        }

        dtype = prepare_numpy_dtype(first_cell.__dict__, list_max_lens=list_max_lens)

        cell_count = len(world.cells)

        array = np.zeros((cell_count,), dtype=dtype).view(np.recarray)

        for n in range(cell_count):
            for k, _ in dtype:
                setattr(array[n], k, getattr(world.cells[n], k))

        return array

    def write(
        self,
        world: World,
        file_name: str,
        time: Optional[float] = None,
        overwrite: bool = False,
        output_count: int = 0,
        **kwargs
    ):
        np.savez(
            check_overwrite(
                ensure_path_and_extension_and_number(file_name, '.npz', output_count),
                overwrite=overwrite,
            ),
            time=time,
            cells=self.output(world),
        )


class CsvOutput(Output):
    """CSV Tabular Output."""

    def output(
        self, world: World, time: Optional[float] = None, **kwargs
    ) -> List[Dict[str, Any]]:
        return [{**cell.__dict__, 'time': time} for cell in world.cells]

    def write(
        self,
        world: World,
        file_name: str,
        time: Optional[float] = None,
        overwrite: bool = False,
        output_count: int = 0,
        **kwargs
    ):
        rows = self.output(world, time=time)
        header = list(sorted(rows[0].keys()))

        with open(
            check_overwrite(
                ensure_path_and_extension_and_number(file_name, '.csv', output_count),
                overwrite=overwrite,
            ),
            'w',
            newline='',
        ) as fp:
            writer = csv.writer(fp)

            writer.writerow(header)

            for row in rows:
                writer.writerow(
                    [row[column] if column in row else '' for column in header]
                )


__all__ = [
    'JsonPickleSerializer',
    'QuickAndDirtyTableDumper',
    'CsvOutput',
]



import importlib
import sys
import types
import warnings
from runpy import run_module

import matplotlib.pyplot
import pytest

from ..cli.simulate import SimulationDuration
from ..random import RRF


def test_random_generator(reset_state):
    fun = RRF.generator.integers(0, 1)
    assert fun() == 0


def test_random_sequence(reset_state):
    seq = RRF.sequence.integers(0, 1)
    assert next(seq) == 0


def test_test_launcher(reset_state, mock_sys_argv):
    with mock_sys_argv(['mock']):
        with pytest.raises(SystemExit):
            run_module('cellsium.tests', run_name='__main__')


def test_launcher(reset_state, mock_sys_argv):
    with mock_sys_argv(['mock', '--help']):
        with pytest.raises(SystemExit):
            run_module('cellsium', run_name='__main__')


def test_launcher_raw(reset_state, mock_sys_argv, tunables, monkeypatch, nop):

    monkeypatch.setattr(matplotlib.pyplot, 'ion', nop)
    monkeypatch.setattr(matplotlib.pyplot, 'show', nop)

    with mock_sys_argv(['mock']):
        with tunables((SimulationDuration, 1.0)):
            with pytest.raises(SystemExit):
                run_module('cellsium', run_name='__main__')


@pytest.mark.parametrize(
    'add_filename,arg',
    [
        (False, None),
        (True, None),
        (True, 'LWPOLYLINE'),
        (True, 'POLYLINE'),
        (True, 'SPLINE'),
    ],
)
def test_dxf_gen(add_filename, arg, tmpdir, mock_sys_argv):

    output_name = tmpdir.join('test.dxf')

    argv = ['mock']
    if add_filename:
        argv.append(str(output_name))
    if arg:
        argv.append(arg)

    with mock_sys_argv(argv):
        with warnings.catch_warnings():
            warnings.simplefilter('ignore')
            run_module('cellsium.tests.generate_dxf_file', run_name='__main__')

    if add_filename:
        assert output_name.isfile()


def test_missing_box2d(
    patch_sys_modules, fix_dynamic_module_reloading_subclass_problems
):
    with fix_dynamic_module_reloading_subclass_problems():
        with patch_sys_modules(remove=['cellsium.simulation.placement*', 'Box2D*']):
            sys.modules['Box2D'] = None
            importlib.import_module('cellsium.simulation.placement')


def test_old_pymunk_mock(
    patch_sys_modules, fix_dynamic_module_reloading_subclass_problems
):
    mock_module = types.ModuleType('pymunkoptions', 'Just a mock pymunkoptions module')
    mock_module.__dict__['options'] = dict()

    with fix_dynamic_module_reloading_subclass_problems():
        with patch_sys_modules(remove=['cellsium.simulation.placement*', 'pymunk*']):
            sys.modules[mock_module.__name__] = mock_module
            importlib.import_module('cellsium.simulation.placement')


def test_patch_sys_modules(patch_sys_modules):
    with patch_sys_modules(remove=None):
        sys.modules['some_non_existing_lib'] = None


def test_instantiate_protocols():
    # function to get the lines covered
    from ..typing import AnyFunction, KwargFunction

    def nop(self, *args, **kwargs):
        pass

    AnyFunction.__init__ = nop
    AnyFunction()(None)

    KwargFunction.__init__ = nop
    KwargFunction()(None, dummy=None)



"""Random number generation infrastructure."""
from typing import Dict, Iterable, Iterator, Type, Union

import numpy as np
from tunable import Tunable

from .typing import AnyFunction, KwargFunction


class Seed(Tunable):
    """Seed for the random number generator"""

    default: int = 1


class RandomNumberGenerator(Tunable):
    """Random number generator to be used"""

    @classmethod
    def available_rngs(cls) -> Dict[str, Type[np.random.BitGenerator]]:
        members = [getattr(np.random, name) for name in dir(np.random)]
        members = [member for member in members if isinstance(member, type)]
        members = [
            member for member in members if np.random.BitGenerator in member.__bases__
        ]
        members = {member.__name__: member for member in members}
        return members

    @classmethod
    def get(cls) -> Type[np.random.BitGenerator]:
        return cls.available_rngs()[cls.value]

    @classmethod
    def test(cls, value: str) -> bool:
        return value in cls.available_rngs()

    default: str = "PCG64"


class RRF:
    """Reproducible random function."""

    seed_value: int = 0

    def __init__(self, mode: str = "callable"):
        assert mode in (
            "callable",
            "iterator",
        )
        self.mode = mode

    def __getattr__(self, item):
        gen = self.spawn_generator()
        func = getattr(gen, item)

        if self.mode == "iterator":

            def _inner(*args, **kwargs):
                while True:
                    yield func(*args, **kwargs)

        else:

            def _inner(*args, **kwargs):
                def _inner_inner():
                    return func(*args, **kwargs)

                return _inner_inner

        return _inner

    @classmethod
    def seed(cls, seed: int = None) -> int:
        """
        Set the seed for the RRF.

        :param seed: Seed
        :return: Seed
        """
        if seed is None:
            seed = Seed.value
        cls.seed_value = seed
        cls.seed_sequence = np.random.SeedSequence(cls.seed_value)
        cls.generator = cls(mode="callable")
        cls.sequence = cls(mode="iterator")
        return seed

    @classmethod
    def spawn_generator(cls) -> np.random.Generator:
        """
        Generates a new np.random.Generator from
        the seed and the configured bitgenerator.

        :return: The Generator instance
        """
        seed = cls.seed_sequence.spawn(1)[0]

        rng = RandomNumberGenerator.get()

        return np.random.Generator(bit_generator=rng(seed=seed))

    @classmethod
    def wrap(cls, sequence: Iterable, func: AnyFunction) -> Iterator:
        """
        Wraps the sequence with the function func so that
        each returned element x becomes func(x).

        :param sequence: Input sequence
        :param func: Function to be called
        :return: Iterator of values
        """
        while True:
            yield func(next(sequence))

    @classmethod
    def compose(cls, func: KwargFunction, **kwargs) -> Iterator:
        """
        Calls a function func with an element \
        of the sequences from the kwargs as kwargs.

        :param func: Function to be called
        :param kwargs: Kwargs of sequences, of which an element each \
        will be used for each function call
        :return: Iterator of values
        """
        while True:
            used_kwargs = {key: next(value) for key, value in kwargs.items()}
            yield func(**used_kwargs)

    @classmethod
    def chain(cls, func: KwargFunction, **kwargs) -> Iterator:
        """
        Calls func with kwargs and yields from it.

        :param func: Function to call
        :param kwargs: Kwargs to pass
        :return: Iterator of values
        """
        while True:
            yield from func(**kwargs)


# seed the RRF so it can be used without explicit initialization
RRF.seed()


def enforce_bounds(
    iterator: Iterator, minimum: float = -np.Inf, maximum: float = np.Inf
) -> Iterator[Union[float, np.ndarray]]:
    """
    Will iter thru an iterator til a value is within bounds.
    For arrays, all values will be considered.

    :param iterator: Iterator
    :param minimum: Minimum value
    :param maximum: Maximum value
    :return: An iterator of values within bounds
    """
    for value in iterator:
        if np.isscalar(value):
            if maximum > value > minimum:
                yield value
        else:
            if ((maximum > np.array(value)) & (np.array(value) > minimum)).all():
                yield value



""" CellSium - _Cell_ _Si_mulator for _micro_fluidic _m_icrocolonies """

__project__ = 'CellSium'

__version__ = '1.0.0'

__author__ = 'Christian C. Sachs'

__copyright__ = '2021, Christian C. Sachs, Forschungszentrum Jülich GmbH'

__citation__ = '''
Sachs CC, Ruzaeva K, Seiffarth J, Wiechert W, Berkels B, Nöh K (2022)
CellSium: versatile cell simulator for microcolony ground truth generation
Bioinformatics Advances, Volume 2, Issue 1, 2022, vbac053, doi: 10.1093/bioadv/vbac053
'''

__banner__ = ''



"""Photorealistic rendered output."""
import os
import warnings
from typing import Dict, Iterator, List, Optional, Set, Tuple

import cv2
import numpy as np
from matplotlib import pyplot
from matplotlib.patches import PathPatch as MatplotlibPathPatch
from matplotlib.path import Path as MatplotlibPath
from roifile import ImagejRoi
from scipy.interpolate import interp1d
from scipy.ndimage.interpolation import rotate
from tifffile import TiffWriter
from tunable import Tunable

from ..model import CellGeometry, WithFluorescence
from ..parameters import Height, Width, um_to_pixel
from ..random import RRF
from ..simulation.simulator import World
from . import (
    Output,
    check_overwrite,
    ensure_path_and_extension,
    ensure_path_and_extension_and_number,
)
from .plot import MicrometerPerCm


def bytescale(image: np.ndarray) -> np.ndarray:
    if image.dtype == np.uint8:
        return image

    low, high = image.min(), image.max()

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        return (255 * ((image - low) / (high - low))).astype(np.uint8)


def noise_attempt(
    times: int = 5, m: int = 10, n: int = 512, r: Optional[Iterator[np.ndarray]] = None
) -> np.ndarray:
    # noinspection PyShadowingNames
    def make_rand(n, r):
        return interp1d(
            np.linspace(0, n, num=n),
            np.array([next(r) for _ in range(n)]),
            kind='cubic',
        )

    # noinspection PyShadowingNames
    def make_square(array):
        return np.tile(array, len(array)).reshape((len(array), len(array)))

    # noinspection PyShadowingNames
    def make_it(n, m, r):
        return make_square(make_rand(m, r)(np.linspace(0, m, n)))

    two_n = 2 * n

    the_sum = make_it(two_n, m, r)

    for _ in range(times):
        the_sum += rotate(make_it(two_n, m, r), angle=360.0 * next(r), reshape=False)

    the_sum /= times

    the_sum = the_sum[n // 2 : two_n - n // 2, n // 2 : two_n - n // 2]
    return the_sum


def gaussian(
    array: np.ndarray, dst: np.ndarray = None, sigma: float = 1.0
) -> np.ndarray:
    return cv2.GaussianBlur(array, (0, 0), sigmaX=um_to_pixel(sigma), dst=dst)


class LuminanceBackground(Tunable):
    """Luminance (lightness) of the background, in 0-1 units."""

    default: float = 0.17


class LuminanceCell(Tunable):
    """Luminance (lightness) of cells, in 0-1 units."""

    default: float = 0.075


def add_if_uneven(value: int, add: int = 1) -> int:
    if (value % 2) == 0:
        return value
    else:
        return value + add


def new_canvas(dtype=np.float32) -> np.ndarray:
    width, height = int(um_to_pixel(Width.value)), int(um_to_pixel(Height.value))
    # make even
    width, height = add_if_uneven(width), add_if_uneven(height)
    canvas = np.zeros((height, width), dtype=dtype)
    return canvas


class OpenCVimshow(Tunable):
    """Show results using OpenCV"""

    default: bool = False


def prepare_patch(coordinates: np.ndarray, **kwargs) -> MatplotlibPathPatch:
    actions = [MatplotlibPath.MOVETO] + [MatplotlibPath.LINETO] * (len(coordinates) - 1)

    if 'closed' in kwargs:
        if kwargs['closed']:
            actions.append(MatplotlibPath.CLOSEPOLY)

            coordinates = np.r_[coordinates, [[0, 0]]]
        del kwargs['closed']

    patch = MatplotlibPathPatch(MatplotlibPath(coordinates, actions), **kwargs)
    return patch


def scale_points_relative(points: np.ndarray, scale_points: float = 1.0) -> np.ndarray:
    if scale_points == 1.0:
        return points

    ma, mi = points.max(axis=0), points.min(axis=0)
    shift = mi + (ma - mi) * 0.5
    points = (scale_points * (points - shift)) + shift
    return points


def scale_points_absolute(points: np.ndarray, delta: float = 0.0) -> np.ndarray:
    if delta == 0.0:
        return points

    ma, mi = points.max(axis=0), points.min(axis=0)
    shift = mi + (ma - mi) * 0.5
    scale_points = (ma - mi + delta) / (ma - mi)
    points = (scale_points * (points - shift)) + shift
    return points


# noinspection PyUnusedLocal
def render_on_canvas_cv2(
    canvas: np.ndarray, array_of_points: np.ndarray, scale_points: float = 1.0, **kwargs
) -> np.ndarray:
    for points in array_of_points:
        points = scale_points_relative(points, scale_points)
        cv2.fillPoly(canvas, points[np.newaxis].astype(np.int32), 1.0)

    return canvas


def render_on_canvas_matplotlib(
    canvas: np.ndarray,
    array_of_points: np.ndarray,
    scale_points: float = 1.0,
    over_sample: int = 1,
) -> np.ndarray:
    interaction_state = pyplot.isinteractive()

    pyplot.ioff()

    dpi = 100.0
    fig = pyplot.figure(
        frameon=False,
        dpi=int(dpi),
        figsize=(
            over_sample * canvas.shape[1] / dpi,
            over_sample * canvas.shape[0] / dpi,
        ),
    )

    ax = fig.add_axes([0, 0, 1, 1])
    for points in array_of_points:
        points = scale_points_relative(points, scale_points)
        ax.add_patch(
            prepare_patch(
                points,
                edgecolor='black',
                facecolor='white',
                closed=True,
                linewidth=over_sample * 0.25,
            )
        )

    plt = ax.imshow(canvas, cmap='gray')

    ax.axis('off')

    plt.axes.get_xaxis().set_visible(False)
    plt.axes.get_yaxis().set_visible(False)

    fig.canvas.draw()

    canvas_data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8).reshape(
        fig.canvas.get_width_height()[::-1] + (3,)
    )[:, :, 0]
    if canvas_data.max() > 0.0:
        canvas_data = canvas_data.astype(np.float32) / canvas_data.max()

    pyplot.close(fig.number)

    if over_sample != 1:
        cv2.resize(
            canvas_data,
            dst=canvas_data,
            dsize=canvas.shape[::-1],
            interpolation=cv2.INTER_AREA,
        )

    if interaction_state:
        pyplot.ion()

    return canvas_data


def get_canvas_points_raw(
    cell: CellGeometry, image_height: Optional[Tuple[int, int]] = None
) -> np.ndarray:
    points = um_to_pixel(cell.points_on_canvas())

    if image_height:
        # flip y, to have (0,0) bottom left
        points[:, 1] = image_height - points[:, 1]

    return points


def get_canvas_points_for_cell(
    cell: CellGeometry, image_height: Optional[Tuple[int, int]] = None
) -> np.ndarray:
    points = get_canvas_points_raw(cell, image_height=image_height)

    if RoiOutputScaleFactor.value != 1.0:
        points = scale_points_relative(points, scale_points=RoiOutputScaleFactor.value)
    if RoiOutputScaleDelta.value != 0.0:
        points = scale_points_absolute(points, delta=RoiOutputScaleDelta.value)

    return points


def cv2_has_write_support(extension: str) -> bool:
    try:
        # Suppress a warning, apparently in some old JPEG2000
        # reading code there were security vulnerabilities
        # Since we only want to write here, it should be fine.
        os.environ["OPENCV_IO_ENABLE_JASPER"] = "1"
        return cv2.haveImageWriter(extension)
    except cv2.error:
        # why does this function throw errors? just return True or False ...
        return False


OpenCV_Acceptable_Write_Extensions = [
    extension
    for extension in (
        '.bmp,.dib,.jpeg,.jpg,.jpe,.jp2,.png,.webp,.pbm,.pgm,'
        '.ppm,.pxm,.pnm,.pfm,.sr,.ras,.tiff,.tif,.exr,.hdr,.pic'
    ).split(',')
    if cv2_has_write_support(extension)
]

if '.png' in OpenCV_Acceptable_Write_Extensions:
    OpenCV_Acceptable_Write_Extensions.remove('.png')
    OpenCV_Acceptable_Write_Extensions = ['.png'] + OpenCV_Acceptable_Write_Extensions


class PlainRenderer(Output):

    write_debug_output = False

    def __init__(self):
        super().__init__()
        self.fig = self.ax = self.imshow_data = None

    @staticmethod
    def new_canvas():
        return new_canvas()

    @staticmethod
    def imwrite(
        name: str,
        img: np.ndarray,
        overwrite: bool = False,
        output_count: Optional[int] = None,
    ) -> bool:
        name = check_overwrite(
            ensure_path_and_extension_and_number(
                name,
                OpenCV_Acceptable_Write_Extensions,
                output_count,
                disable_individual=not output_count,
            ),
            overwrite=overwrite,
        )

        return cv2.imwrite(name, img)

    def debug_output(self, name: str, array: np.ndarray) -> None:
        if not self.write_debug_output:
            return

        self.imwrite("render-%s.png" % (name,), bytescale(array), overwrite=True)

    @staticmethod
    def render_cells(
        canvas: np.ndarray, array_of_points: np.ndarray, fast: bool = False
    ) -> np.ndarray:
        if fast:
            canvas = render_on_canvas_cv2(canvas, array_of_points)
        else:
            canvas = render_on_canvas_matplotlib(canvas, array_of_points)
        return canvas

    def output(self, world: World, **kwargs) -> np.ndarray:
        canvas = self.new_canvas()

        array_of_points = [
            get_canvas_points_raw(cell, canvas.shape[0]) for cell in world.cells
        ]

        canvas = self.render_cells(canvas, array_of_points)

        self.debug_output('raw-cells', canvas)

        return canvas

    @staticmethod
    def convert(image: np.ndarray, max_value: int = 255) -> np.ndarray:
        return (np.clip(image, 0, 1) * max_value).astype(np.uint8)

    def write(
        self,
        world: World,
        file_name: str,
        overwrite: bool = False,
        output_count: int = 0,
        **kwargs
    ) -> None:
        self.imwrite(
            file_name,
            self.convert(self.output(world)),
            overwrite=overwrite,
            output_count=output_count,
        )

    def display(self, world: World, **kwargs) -> None:

        image = self.output(world)

        if OpenCVimshow.value:
            cv2.imshow(self.__class__.__name__, self.convert(image))
            cv2.waitKey()
        else:
            from matplotlib import pyplot

            pyplot.ion()

            if getattr(self, 'fig', None) is None:
                self.fig = pyplot.figure(
                    figsize=(
                        Width.value / MicrometerPerCm.value / 2.51,
                        Height.value / MicrometerPerCm.value / 2.51,
                    )
                )

            if getattr(self, 'ax', None) is None:
                self.ax = self.fig.add_subplot(111)

            if getattr(self, 'imshow_data', None):
                self.imshow_data.set_data(image)
            else:
                self.ax.clear()
                self.imshow_data = self.ax.imshow(image, cmap='gray')

                with warnings.catch_warnings():
                    warnings.simplefilter('ignore')
                    pyplot.tight_layout()

                pyplot.show()

            self.fig.canvas.draw()
            self.fig.canvas.flush_events()


class FluorescenceEmitterKernelSizeW(Tunable):
    """Kernel size for fluorescence emitter placement, width"""

    default: float = 17


class FluorescenceEmitterKernelSizeH(Tunable):
    """Kernel size for fluorescence emitter placement, height"""

    default: float = 17


class FluorescenceEmitterGaussianW(Tunable):
    """Fluorescence emitter placement Gaussian sigma, horizontal"""

    default: float = 2.5


class FluorescenceEmitterGaussianH(Tunable):
    """Fluorescence emitter placement Gaussian sigma, vertical"""

    default: float = 2.5


class FluorescenceNoiseMean(Tunable):
    """Fluorescence noise mean"""

    default: float = 0.0


class FluorescenceNoiseStd(Tunable):
    """Fluorescence noise standard deviation"""

    default: float = 0.15


class FluorescenceRatioBackground(Tunable):
    """Fluorescence background ratio"""

    default: float = 1.0


class FluorescenceCellSizeFactor(Tunable):
    """Fluorescence cell size to fluorescence factor"""

    default: float = 500.0


class FluorescenceRenderer(PlainRenderer):

    channel: int = 0

    def __init__(self):
        super().__init__()
        canvas = self.new_canvas()

        self.rng = RRF.spawn_generator()

        self.random_noise = RRF.sequence.normal(
            FluorescenceNoiseMean.value,
            FluorescenceNoiseStd.value,
            canvas.shape,
        )

    def output(self, world: World, **kwargs) -> np.ndarray:
        canvas = self.new_canvas()

        int_background = FluorescenceRatioBackground.value

        emitter_size = (
            FluorescenceEmitterKernelSizeW.value,
            FluorescenceEmitterKernelSizeH.value,
        )

        emitter = np.zeros(emitter_size)
        emitter[emitter_size[0] // 2, emitter_size[1] // 2] = 1.0

        emitter = cv2.GaussianBlur(
            emitter,
            emitter_size,
            sigmaX=FluorescenceEmitterGaussianW.value,
            sigmaY=FluorescenceEmitterGaussianH.value,
        )

        self.debug_output('fluorescence-emitter', emitter)

        p_canvas = np.pad(canvas, emitter_size, mode='reflect')

        for cell in world.cells:
            points = um_to_pixel(cell.points_on_canvas())
            #
            points[:, 1] = canvas.shape[0] - points[:, 1]

            pts = points[np.newaxis].astype(np.int32)

            # Skip cells which (partly) lie outside of the image
            # TODO proper handling, so that parts of cells
            #  poking into the image are still properly handled
            if (points[:, 0].min() < 0 or points[:, 0].max() > canvas.shape[1]) or (
                points[:, 1].min() < 0 or points[:, 1].max() > canvas.shape[0]
            ):
                continue

            if isinstance(cell, WithFluorescence):
                if self.channel < len(cell.fluorescences):
                    brightness = cell.fluorescences[self.channel] * int_background
                else:
                    brightness = 0

            else:
                brightness = 0

            # TODO normalize by µm^2
            int_countdown = brightness * (
                cv2.contourArea(pts) / FluorescenceCellSizeFactor.value
            )

            points_x_min, points_x_max = points[:, 0].min(), points[:, 0].max()
            points_y_min, points_y_max = points[:, 1].min(), points[:, 1].max()

            while int_countdown > 0:
                # TODO needs tests
                x = self.rng.uniform(points_x_min, points_x_max)
                y = self.rng.uniform(points_y_min, points_y_max)

                if cv2.pointPolygonTest(pts, (x, y), measureDist=False) > 0.0:
                    target = p_canvas[
                        int(y)
                        + emitter_size[1] // 2 : int(y)
                        + 3 * emitter_size[1] // 2,
                        int(x)
                        + emitter_size[0] // 2 : int(x)
                        + 3 * emitter_size[0] // 2,
                    ]
                    target += emitter[: target.shape[0], : target.shape[1]]
                    int_countdown -= 1

        canvas = p_canvas[
            emitter_size[0] : -emitter_size[0], emitter_size[1] : -emitter_size[1]
        ]

        self.debug_output('raw-fluorescence-cells', canvas)

        noise = int_background * np.abs(next(self.random_noise))

        self.debug_output('raw-fluorescence-noise', noise)

        canvas += noise

        self.debug_output('fluorescence', canvas)

        return canvas


class PhaseContrastRenderer(PlainRenderer):
    def output(self, world: World, **kwargs) -> np.ndarray:
        cell_canvas = super().output(world)

        background = LuminanceBackground.value * np.ones_like(cell_canvas)

        self.debug_output('pc-background', background)

        cell_halo = 0.5 * gaussian(cell_canvas, sigma=0.75)

        background_w_halo = background + cell_halo

        self.debug_output('pc-halo-background', background_w_halo)

        cell_canvas = gaussian(cell_canvas, sigma=0.05)

        self.debug_output('pc-blurred-cells', cell_canvas)

        halo_glow_in_cells = (
            0.4
            * gaussian(cell_canvas, 0.1)
            * gaussian(cell_halo * (1 - cell_canvas), sigma=0.05)
        )

        self.debug_output('pc-blur-in-cells', halo_glow_in_cells)

        result = (
            background_w_halo * (1 - cell_canvas) + cell_canvas * LuminanceCell.value
        ) + halo_glow_in_cells

        self.debug_output('pc-unblurred-result', result)

        result = gaussian(result, dst=result, sigma=0.075)

        self.debug_output('pc-result', result)

        return result


class UnevenIlluminationAdditiveFactor(Tunable):
    """Additive component for the uneven illumination"""

    default: float = 0.02


class UnevenIlluminationMultiplicativeFactor(Tunable):
    """Multiplicative factor for the uneven illumination"""

    # default: float = 0.05
    default: float = 0.25


class UnevenIlluminationPhaseContrast(PhaseContrastRenderer):
    def __init__(self):
        super().__init__()

        self.random_complex_noise = RRF.sequence.uniform(0, 1)
        self.uneven_illumination = None
        self.create_uneven_illumination()

    def new_uneven_illumination(self) -> np.ndarray:
        empty = self.new_canvas()
        return noise_attempt(
            times=5, m=10, n=max(empty.shape), r=self.random_complex_noise
        )[: empty.shape[0], : empty.shape[1]]

    def create_uneven_illumination(self) -> np.ndarray:
        self.uneven_illumination = self.new_uneven_illumination()

    def output(self, world: World, **kwargs) -> np.ndarray:
        canvas = super().output(world)

        self.debug_output('uneven-illumination', self.uneven_illumination)

        canvas = (
            canvas
            * (
                1.0
                + UnevenIlluminationMultiplicativeFactor.value
                * self.uneven_illumination
            )
        ) + UnevenIlluminationAdditiveFactor.value * self.uneven_illumination

        self.debug_output('pc-with-uneven-illumination', canvas)

        return canvas


class NoisyUnevenIlluminationPhaseContrast(UnevenIlluminationPhaseContrast):
    def __init__(self):
        super().__init__()

        empty = self.new_canvas()

        self.product_noise = RRF.sequence.normal(1.0, 0.002, empty.shape)
        self.sum_noise = RRF.sequence.normal(0.0, 0.002, empty.shape)

    def output(self, world: World, **kwargs) -> np.ndarray:
        canvas = super().output(world)

        product_noise = next(self.product_noise)
        sum_noise = next(self.sum_noise)

        self.debug_output('product_noise', product_noise)
        self.debug_output('sum_noise', sum_noise)

        canvas = canvas * product_noise + sum_noise

        self.debug_output('pc-uneven-w-noise', canvas)

        return canvas


class RoiOutputScaleFactor(Tunable):
    """Scale factor for the ROI output"""

    default: float = 1.0


class RoiOutputScaleDelta(Tunable):
    """Scale delta for the ROI output"""

    default: float = -4.0


def collect_subclasses(start: type) -> Set[type]:
    collector = set()

    def _recurse(class_):
        collector.add(class_)
        for subclass_ in class_.__subclasses__():
            _recurse(subclass_)

    _recurse(start)
    return collector


class RenderChannels(Tunable):
    """Channels to render (i.e. output classes which produce rendered images)"""

    default = 'NoisyUnevenIlluminationPhaseContrast'
    # default = 'NoisyUnevenIlluminationPhaseContrast, FluorescenceRenderer'

    @staticmethod
    def get_mapping() -> Dict[str, type]:
        return {class_.__name__: class_ for class_ in collect_subclasses(PlainRenderer)}

    @classmethod
    def test(cls, value: str) -> bool:
        mapping = cls.get_mapping()
        for class_ in value.split(","):
            if class_.strip() not in mapping:
                return False
        return True

    @classmethod
    def instantiate(cls) -> PlainRenderer:
        mapping = cls.get_mapping()
        return [mapping[class_.strip()]() for class_ in cls.value.split(',')]


class TiffOutput(Output):
    output_type: np.dtype = np.uint8

    def __init__(self):
        self.channels = RenderChannels.instantiate()
        self.images = []
        self.current = -1
        self.file_name = None
        self.rois = []

    def output(self, world: World, **kwargs) -> List[np.ndarray]:
        return [c.output(world) for c in self.channels]

    def __del__(self):
        if not self.images:
            return

        stack = []
        for stacklet in self.images:
            stacklet = np.concatenate([image[np.newaxis] for image in stacklet], axis=0)
            stack.append(stacklet[np.newaxis, np.newaxis])

        result = np.concatenate(stack)

        if self.output_type in (np.uint8, np.uint16):

            for c in range(result.shape[2]):
                result[:, 0, c, :, :] -= result[:, 0, c, :, :].min()
                result[:, 0, c, :, :] /= result[:, 0, c, :, :].max()

            result *= 2 ** (8 * np.dtype(self.output_type).itemsize) - 1

        result = result.astype(self.output_type)

        binary_rois = [ImagejRoi.frompoints(**roi).tobytes() for roi in self.rois]

        with TiffWriter(
            ensure_path_and_extension(self.file_name, '.tif'), imagej=True
        ) as writer:
            writer.save(
                result,
                resolution=(um_to_pixel(1.0), um_to_pixel(1.0)),
                metadata=dict(unit='um', Overlays=binary_rois),
            )

    def write(self, world: World, file_name: str, **kwargs) -> None:
        self.file_name = file_name

        self.current += 1

        stacklet = self.output(world)

        self.images.append(stacklet)

        for idx, cell in enumerate(world.cells):
            points = get_canvas_points_for_cell(cell, image_height=stacklet[0].shape[0])

            if len(self.channels) > 1:
                self.rois.append(
                    dict(points=points, t=self.current, position=-1, index=idx)
                )
            else:
                self.rois.append(
                    dict(points=points, t=-1, position=self.current, index=idx)
                )


__all__ = [
    'RenderChannels',
    'get_canvas_points_for_cell',
    'new_canvas',
    'PlainRenderer',
    'FluorescenceRenderer',
    'PhaseContrastRenderer',
    'UnevenIlluminationPhaseContrast',
    'NoisyUnevenIlluminationPhaseContrast',
    'TiffOutput',
]



"""Output as SVG vector images."""
# noinspection PyPep8Naming
import xml.etree.ElementTree as ET

import numpy as np
from tunable import Tunable

from ..parameters import Height, Width
from ..simulation.simulator import World
from . import Output, check_overwrite, ensure_path_and_extension_and_number


class MicrometerPerCm(Tunable):
    """Calibration for outputs, micrometer per centimeter"""

    default: float = 2.5


class SvgRenderer(Output):
    @staticmethod
    def create_xml():
        root = ET.Element('svg')

        xml = ET.ElementTree(root)

        root.attrib['width'] = str(Width.value)
        root.attrib['height'] = str(Height.value)

        defs = ET.SubElement(root, 'defs')
        style = ET.SubElement(defs, 'style')

        style.attrib['type'] = 'text/css'
        style.text = """
        .cell {
            stroke: #005b82;
            stroke-width: 0.1px;
            fill: none;
        }
        .boundary {
            stroke: #005b82;
            stroke-width: 0.1px;
            fill: none;
        }
        """

        return xml

    def __init__(self):
        super().__init__()

        self.xml = self.create_xml()

        self.group_boundary = ET.SubElement(self.xml.getroot(), 'g')
        self.group_boundary.attrib['id'] = 'boundaries'
        self.group_cells_all = ET.SubElement(self.xml.getroot(), 'g')
        self.group_boundary.attrib['id'] = 'cells'

        self.boundaries_written = False

    @staticmethod
    def points_to_path(points: np.ndarray) -> str:
        return 'M' + (''.join('L%f %f' % tuple(point) for point in points))[1:] + 'Z'

    def output(self, world: World, **kwargs) -> None:

        if not self.boundaries_written:

            for boundary in world.boundaries:
                boundary_element = ET.SubElement(self.group_boundary, 'path')
                boundary_element.attrib['class'] = 'boundary'
                boundary_element.attrib['d'] = self.points_to_path(boundary)

            self.boundaries_written = True

        # frame
        group = ET.SubElement(self.group_cells_all, 'g')
        group.attrib['id'] = 'frameXXX'

        for cell in world.cells:
            cell_path = ET.SubElement(group, 'path')
            cell_path.attrib['class'] = 'cell'
            cell_path.attrib['d'] = self.points_to_path(cell.points_on_canvas())

    def write(
        self,
        world: World,
        file_name: str,
        overwrite: bool = False,
        output_count: int = 0,
        **kwargs
    ) -> None:
        self.output(world)
        self.xml.write(
            check_overwrite(
                ensure_path_and_extension_and_number(file_name, '.svg', output_count),
                overwrite=overwrite,
            ),
            encoding='utf-8',
        )

    def display(self, world: World, **kwargs) -> None:
        raise RuntimeError('Unsupported')


__all__ = ['SvgRenderer']
